# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'galacteek/ui/galacteek.ui'
#
# Created by: PyQt5 UI code generator 5.10
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_GalacteekWindow(object):
    def setupUi(self, GalacteekWindow):
        GalacteekWindow.setObjectName("GalacteekWindow")
        GalacteekWindow.resize(1038, 501)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(GalacteekWindow.sizePolicy().hasHeightForWidth())
        GalacteekWindow.setSizePolicy(sizePolicy)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/share/icons/galacteek.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        GalacteekWindow.setWindowIcon(icon)
        self.centralwidget = QtWidgets.QWidget(GalacteekWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.verticalLayout_3 = QtWidgets.QVBoxLayout(self.centralwidget)
        self.verticalLayout_3.setObjectName("verticalLayout_3")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.tabWidget = QtWidgets.QTabWidget(self.centralwidget)
        self.tabWidget.setObjectName("tabWidget")
        self.tabBlank = QtWidgets.QWidget()
        self.tabBlank.setObjectName("tabBlank")
        self.verticalLayout_6 = QtWidgets.QVBoxLayout(self.tabBlank)
        self.verticalLayout_6.setObjectName("verticalLayout_6")
        self.tabWidget.addTab(self.tabBlank, "")
        self.verticalLayout.addWidget(self.tabWidget)
        self.verticalLayout_3.addLayout(self.verticalLayout)
        GalacteekWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(GalacteekWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 1038, 20))
        self.menubar.setDefaultUp(True)
        self.menubar.setObjectName("menubar")
        GalacteekWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(GalacteekWindow)
        self.statusbar.setObjectName("statusbar")
        GalacteekWindow.setStatusBar(self.statusbar)
        self.actionQuit = QtWidgets.QAction(GalacteekWindow)
        self.actionQuit.setPriority(QtWidgets.QAction.LowPriority)
        self.actionQuit.setObjectName("actionQuit")
        self.actionAddProfile = QtWidgets.QAction(GalacteekWindow)
        self.actionAddProfile.setObjectName("actionAddProfile")
        self.actionShowPeersInformation = QtWidgets.QAction(GalacteekWindow)
        self.actionShowPeersInformation.setObjectName("actionShowPeersInformation")
        self.actionShowIpfsRefs = QtWidgets.QAction(GalacteekWindow)
        self.actionShowIpfsRefs.setObjectName("actionShowIpfsRefs")
        self.actionSettings = QtWidgets.QAction(GalacteekWindow)
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/share/icons/settings.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionSettings.setIcon(icon1)
        self.actionSettings.setObjectName("actionSettings")
        self.actionAboutGalacteek = QtWidgets.QAction(GalacteekWindow)
        self.actionAboutGalacteek.setObjectName("actionAboutGalacteek")
        self.actionCloseAllTabs = QtWidgets.QAction(GalacteekWindow)
        self.actionCloseAllTabs.setPriority(QtWidgets.QAction.HighPriority)
        self.actionCloseAllTabs.setObjectName("actionCloseAllTabs")
        self.actionNew_Profile = QtWidgets.QAction(GalacteekWindow)
        self.actionNew_Profile.setObjectName("actionNew_Profile")
        self.actionDonate = QtWidgets.QAction(GalacteekWindow)
        self.actionDonate.setObjectName("actionDonate")
        self.actionEvent_log = QtWidgets.QAction(GalacteekWindow)
        self.actionEvent_log.setObjectName("actionEvent_log")
        self.actionBrowse = QtWidgets.QAction(GalacteekWindow)
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/share/icons/ipfs-logo-128-white-outline.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.actionBrowse.setIcon(icon2)
        self.actionBrowse.setObjectName("actionBrowse")
        self.actionExitApp = QtWidgets.QAction(GalacteekWindow)
        self.actionExitApp.setObjectName("actionExitApp")

        self.retranslateUi(GalacteekWindow)
        self.tabWidget.setCurrentIndex(0)
        QtCore.QMetaObject.connectSlotsByName(GalacteekWindow)

    def retranslateUi(self, GalacteekWindow):
        _translate = QtCore.QCoreApplication.translate
        GalacteekWindow.setWindowTitle(_translate("GalacteekWindow", "Galacteek"))
        self.actionQuit.setText(_translate("GalacteekWindow", "Quit"))
        self.actionQuit.setShortcut(_translate("GalacteekWindow", "Ctrl+Q"))
        self.actionAddProfile.setText(_translate("GalacteekWindow", "Add Profile"))
        self.actionShowPeersInformation.setText(_translate("GalacteekWindow", "Show peers information"))
        self.actionShowIpfsRefs.setText(_translate("GalacteekWindow", "Show refs"))
        self.actionSettings.setText(_translate("GalacteekWindow", "Settings"))
        self.actionAboutGalacteek.setText(_translate("GalacteekWindow", "About"))
        self.actionCloseAllTabs.setText(_translate("GalacteekWindow", "Close all tabs"))
        self.actionNew_Profile.setText(_translate("GalacteekWindow", "New Profile"))
        self.actionDonate.setText(_translate("GalacteekWindow", "Donate"))
        self.actionEvent_log.setText(_translate("GalacteekWindow", "Event log"))
        self.actionBrowse.setText(_translate("GalacteekWindow", "Browse"))
        self.actionExitApp.setText(_translate("GalacteekWindow", "Quit"))

from . import galacteek_rc



import re
import asyncio

from PyQt5.QtWidgets import QWidget
from PyQt5.QtWidgets import QToolButton
from PyQt5.QtWidgets import QLineEdit
from PyQt5.QtWidgets import QMenu
from PyQt5.QtWidgets import QAction
from PyQt5.QtWidgets import QVBoxLayout
from PyQt5.QtWidgets import QWidgetAction
from PyQt5.QtWidgets import QLabel

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtCore import Qt
from PyQt5.QtCore import QPoint

from PyQt5.QtGui import QImage
from PyQt5.QtGui import QPixmap

from galacteek.ipfs.wrappers import ipfsOp
from galacteek import ensure

from .helpers import getIcon
from .helpers import getIconFromIpfs
from .helpers import disconnectSig
from .helpers import sizeFormat
from .i18n import iNoTitle
from .i18n import iCancel
from .i18n import iUnknown
from .i18n import iHashmarksLibraryCountAvailable
from .i18n import iLocalHashmarksCount


class GalacteekTab(QWidget):
    def __init__(self, gWindow, **kw):
        super(GalacteekTab, self).__init__(gWindow)
        self.vLayout = QVBoxLayout(self)
        self.setLayout(self.vLayout)

        self.gWindow = gWindow
        self.setAttribute(Qt.WA_DeleteOnClose)

    def addToLayout(self, widget):
        self.vLayout.addWidget(widget)

    def onClose(self):
        return True

    @ipfsOp
    async def initialize(self, op):
        pass

    @property
    def app(self):
        return self.gWindow.app

    @property
    def loop(self):
        return self.app.loop

    @property
    def profile(self):
        return self.app.ipfsCtx.currentProfile


class PopupToolButton(QToolButton):
    def __init__(self, icon=None, parent=None, menu=None,
                 mode=QToolButton.MenuButtonPopup):
        super(PopupToolButton, self).__init__(parent)

        self.menu = menu if menu else QMenu()
        self.setPopupMode(mode)
        self.setMenu(self.menu)

        if icon:
            self.setIcon(icon)


class _HashmarksCommon:
    def makeAction(self, path, mark):
        tLenMax = 64
        title = mark['metadata'].get('title', iNoTitle())
        icon = mark.get('icon', None)
        fullTitle = title

        if len(title) > tLenMax:
            title = '{0} ...'.format(title[0:tLenMax])

        action = QAction(title, self)
        action.setToolTip(fullTitle)
        action.setData({
            'path': path,
            'mark': mark,
            'iconcid': icon
        })

        if not icon:
            # Default icon
            action.setIcon(getIcon('ipfs-logo-128-white-outline.png'))

        return action

    @ipfsOp
    async def loadMarkIcon(self, ipfsop, action, iconCid):
        icon = await getIconFromIpfs(ipfsop, iconCid)
        if icon:
            action.setIcon(icon)


class HashmarkMgrButton(PopupToolButton, _HashmarksCommon):
    hashmarkClicked = pyqtSignal(str, str)

    def __init__(self, marks, iconFile='hashmarks.png',
                 maxItemsPerCategory=128, parent=None):
        super(HashmarkMgrButton, self).__init__(parent=parent,
                                                mode=QToolButton.InstantPopup)

        self.setObjectName('hashmarksMgrButton')
        self.menu.setObjectName('hashmarksMgrMenu')
        self.hCount = 0
        self.marks = marks
        self.cMenus = {}
        self.maxItemsPerCategory = maxItemsPerCategory
        self.setIcon(getIcon(iconFile))

        disconnectSig(self.marks.changed, self.onChanged)

        self.marks.changed.connect(self.onChanged)

    def onChanged(self):
        self.updateMenu()

    async def updateIcons(self):
        for mName, menu in self.cMenus.items():
            await asyncio.sleep(0)

            for action in menu.actions():
                data = action.data()

                if data['iconcid']:
                    ensure(self.loadMarkIcon(action, data['iconcid']))

    def updateMenu(self):
        self.hCount = 0
        categories = self.marks.getCategories()

        for category in categories:
            marks = self.marks.getCategoryMarks(category)
            mItems = marks.items()

            if len(mItems) not in range(1, self.maxItemsPerCategory):
                continue

            if category not in self.cMenus:
                self.cMenus[category] = QMenu(category)
                self.cMenus[category].triggered.connect(self.linkActivated)
                self.cMenus[category].setObjectName('hashmarksMgrMenu')
                self.menu.addMenu(self.cMenus[category])

            menu = self.cMenus[category]
            menu.setIcon(getIcon('stroke-cube.png'))

            def exists(path):
                for action in menu.actions():
                    if action.data()['path'] == path:
                        return action

            if len(mItems) in range(1, self.maxItemsPerCategory):
                for path, mark in mItems:
                    self.hCount += 1

                    if exists(path):
                        continue

                    menu.addAction(self.makeAction(path, mark))
            else:
                menu.hide()

        self.setToolTip(iLocalHashmarksCount(self.hCount))

    def linkActivated(self, action):
        data = action.data()
        path, mark = data['path'], data['mark']

        if mark:
            if 'metadata' not in mark:
                return
            self.hashmarkClicked.emit(
                path,
                mark['metadata']['title']
            )


class HashmarksLibraryButton(PopupToolButton, _HashmarksCommon):
    hashmarkClicked = pyqtSignal(str, str)

    def __init__(self, iconFile='hashmarks-library.png',
                 maxItemsPerCategory=32, parent=None):
        super(HashmarksLibraryButton, self).__init__(
            parent=parent, mode=QToolButton.InstantPopup)

        self.cMenus = {}
        self.hCount = 0
        self.maxItemsPerCategory = maxItemsPerCategory
        self.setIcon(getIcon(iconFile))
        self.setObjectName('hashmarksLibraryButton')
        self.menu.setObjectName('hashmarksLibraryMenu')
        self.searchMenu = None
        self.addSearchMenu()

    def addSearchMenu(self):
        if self.searchMenu is not None:
            return

        self.searchMenu = QMenu('Search')
        self.menu.addMenu(self.searchMenu)

        self.searchLine = QLineEdit()
        self.searchLine.returnPressed.connect(self.onSearch)
        self.searchWAction = QWidgetAction(self)
        self.searchWAction.setDefaultWidget(self.searchLine)
        self.searchMenu.addAction(self.searchWAction)
        self.searchMenu.setEnabled(False)

    def onSearch(self):
        pos = self.searchLine.mapToGlobal(QPoint(0, 0))
        text = self.searchLine.text()

        resultsMenu = QMenu()
        resultsMenu.triggered.connect(
            lambda action: self.linkActivated(
                action, closeMenu=True))
        self.searchTextInMenu(self.menu, text, resultsMenu)
        resultsMenu.exec(pos)

    def searchTextInMenu(self, menu, text, rMenu):
        for action in menu.actions():
            menu = action.menu()
            if menu:
                self.searchTextInMenu(menu, text, rMenu)
            else:
                data = action.data()
                if not isinstance(data, dict):
                    continue

                path = data['path']
                mark = data['mark']

                maTitle = re.search(text, mark['metadata']['title'],
                                    re.IGNORECASE)
                maDesc = re.search(text, mark['metadata']['description'],
                                   re.IGNORECASE)
                if maTitle or maDesc:
                    rMenu.addAction(self.makeAction(path, mark))

    def updateMenu(self, ipfsMarks):
        categories = ipfsMarks.getCategories()

        for category in categories:
            marks = ipfsMarks.getCategoryMarks(category)
            mItems = marks.items()

            if len(mItems) not in range(1, self.maxItemsPerCategory):
                continue

            if category not in self.cMenus:
                self.cMenus[category] = QMenu(category)
                self.cMenus[category].triggered.connect(self.linkActivated)
                self.cMenus[category].setObjectName('hashmarksLibraryMenu')
                self.menu.addMenu(self.cMenus[category])

            menu = self.cMenus[category]

            menu.setIcon(getIcon('stroke-cube.png'))

            def exists(path):
                for action in menu.actions():
                    if action.data()['path'] == path:
                        return action

            for path, mark in mItems:
                self.hCount += 1
                if exists(path):
                    continue

                action = self.makeAction(path, mark)
                menu.addAction(action)

        self.setToolTip(iHashmarksLibraryCountAvailable(self.hCount))
        if self.hCount > 0:
            self.searchMenu.setEnabled(True)

    def linkActivated(self, action, closeMenu=False):
        data = action.data()
        path, mark = data['path'], data['mark']

        if mark:
            if 'metadata' not in mark:
                return

            self.hashmarkClicked.emit(
                path,
                mark['metadata']['title']
            )

            if closeMenu:
                self.menu.hide()


class ImageWidget(QLabel):
    """
    Displays an image stored in IPFS into a QLabel
    """

    def __init__(self, scaleWidth=64, parent=None):
        super(ImageWidget, self).__init__(parent)

        self._imgPath = None
        self._scaleWidth = scaleWidth

    @property
    def imgPath(self):
        return self._imgPath

    @ipfsOp
    async def load(self, ipfsop, path):
        try:
            imgData = await ipfsop.waitFor(
                ipfsop.client.cat(path), 8
            )

            if not imgData:
                raise Exception('Failed to load image')

            img1 = QImage()
            img1.loadFromData(imgData)
            img = img1.scaledToWidth(self._scaleWidth)

            self.setPixmap(QPixmap.fromImage(img))
            self._imgPath = path
            return True
        except Exception:
            return False


class DownloadProgressButton(PopupToolButton):
    downloadProgress = pyqtSignal(int)
    downloadFinished = pyqtSignal()
    cancelled = pyqtSignal()

    def __init__(self, path, stat, parent=None):
        super(DownloadProgressButton, self).__init__(
            icon=getIcon('download.png'),
            parent=parent, mode=QToolButton.InstantPopup)
        self.downloadProgress.connect(self.onProgress)
        self.downloadFinished.connect(self.onFinished)
        self.path = path
        self.stat = stat
        self.readBytes = 0
        self.task = None

        self.menu.addAction(iCancel(), self.onCancel)

    def onCancel(self):
        if self.task:
            try:
                self.task.cancel()
            except:
                pass

            self.cancelled.emit()

    def onProgress(self, read):
        self.readBytes = read
        size = sizeFormat(self.stat.get('CumulativeSize')) if self.stat else \
            iUnknown()

        self.setToolTip('{path} (size: {size}): downloaded {dl}'.format(
            path=self.path,
            size=size,
            dl=sizeFormat(self.readBytes))
        )

    def onFinished(self):
        self.setToolTip('{path}: download finished'.format(path=self.path))



import os
import os.path
import asyncio
import aioipfs

from PyQt5.QtWidgets import QApplication
from PyQt5.QtCore import QCoreApplication
from PyQt5.QtCore import QObject
from PyQt5.QtCore import pyqtSignal

from galacteek import log
from galacteek import logUser
from galacteek import ensure

from galacteek.dweb.page import PDFViewerPage
from galacteek.dweb.page import WebTab
from galacteek.dweb.page import DWebView

from galacteek.ipfs import ipfsOp
from galacteek.ipfs.mimetype import MIMEType
from galacteek.ipfs.mimetype import detectMimeType

from galacteek.ipfs.cidhelpers import IPFSPath
from galacteek.ipfs.cidhelpers import shortPathRepr

from galacteek.crypto.qrcode import IPFSQrDecoder

from . import ipfsview
from .helpers import getMimeIcon
from .helpers import messageBox
from .imgview import ImageViewerTab


def iResourceCannotOpen(path):
    return QCoreApplication.translate(
        'resourceOpener',
        '{}: cannot determine resource type or timeout occured '
        '(check connectivity)').format(
            path)


class IPFSResourceOpener(QObject):
    objectOpened = pyqtSignal(IPFSPath)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.app = QApplication.instance()
        self.setObjectName('resourceOpener')

    @ipfsOp
    async def open(self, ipfsop, pathRef, mimeType=None):
        """
        Open the resource referenced by rscPath according
        to its MIME type

        :param pathRef: IPFS path (can be str or IPFSPath)
        :param MIMEType mimeType: MIME type
        """

        ipfsPath = None
        if isinstance(pathRef, IPFSPath):
            ipfsPath = pathRef
        elif isinstance(pathRef, str):
            ipfsPath = IPFSPath(pathRef)
        else:
            raise ValueError('Invalid input value')

        if not ipfsPath.valid:
            return False

        rscPath = ipfsPath.objPath

        if self.app.mainWindow.pinAllGlobalChecked and not ipfsPath.isIpns:
            ensure(ipfsop.ctx.pinner.queue(rscPath, False,
                                           None,
                                           qname='default'))

        rscShortName = shortPathRepr(rscPath)

        if ipfsPath.isIpfs:
            # Try to reuse metadata from the multihash store
            rscMeta = await self.app.multihashDb.get(rscPath)
            if rscMeta:
                value = rscMeta.get('mimetype')
                if value:
                    mimeType = MIMEType(value)

        if mimeType is None:
            mimeType = await detectMimeType(rscPath)

        if mimeType and mimeType.valid:
            logUser.info('{path} ({type}): opening'.format(
                path=rscPath, type=str(mimeType)))
        else:
            return messageBox(iResourceCannotOpen(rscPath))

        if mimeType.isText:
            tab = ipfsview.TextViewerTab()
            tab.show(rscPath)
            self.objectOpened.emit(ipfsPath)
            return self.app.mainWindow.registerTab(
                tab,
                rscShortName,
                icon=getMimeIcon('text/x-generic'),
                tooltip=rscPath,
                current=True
            )

        if mimeType.isImage:
            tab = ImageViewerTab(rscPath, self.app.mainWindow)
            self.objectOpened.emit(ipfsPath)
            return self.app.mainWindow.registerTab(
                tab,
                rscShortName,
                icon=getMimeIcon('image/x-generic'),
                tooltip=rscPath,
                current=True
            )

        if mimeType.isVideo or mimeType.isAudio:
            tab = self.app.mainWindow.addMediaPlayerTab()
            if tab:
                tab.playFromPath(rscPath)
            return

        if mimeType == 'application/pdf':
            return await self.openWithSystemDefault(rscPath)

        if mimeType == 'application/pdf' and 0:  # not usable yet
            tab = WebTab(self.app.mainWindow)
            tab.attach(
                DWebView(page=PDFViewerPage(rscPath))
            )
            self.objectOpened.emit(ipfsPath)
            return self.app.mainWindow.registerTab(
                tab,
                rscShortName,
                icon=getMimeIcon('application/pdf'),
                tooltip=rscPath,
                current=True
            )

        if mimeType.isDir or ipfsPath.isIpns or mimeType.isHtml:
            self.objectOpened.emit(ipfsPath)
            return self.app.mainWindow.addBrowserTab().browseFsPath(ipfsPath)

        logUser.info('{path} ({type}): unhandled resource type'.format(
            path=rscPath, type=str(mimeType)))

    @ipfsOp
    async def openWithExternal(self, ipfsop, rscPath, progArgs):
        filePath = os.path.join(self.app.tempDir.path(),
                                os.path.basename(rscPath))

        if not os.path.exists(filePath):
            try:
                await ipfsop.client.get(rscPath,
                                        dstdir=self.app.tempDir.path())
            except aioipfs.APIError as err:
                log.debug(err.message)
                return

        if not os.path.isfile(filePath):
            # Bummer
            return

        for idx, value in enumerate(progArgs):
            if value == '%f':
                progArgs[idx] = filePath

        try:
            proc = await asyncio.create_subprocess_exec(
                *progArgs,
                stdout=asyncio.subprocess.PIPE
            )
            stdout, stderr = await proc.communicate()
        except BaseException as err:
            os.unlink(filePath)
            log.debug(str(err))
        else:
            os.unlink(filePath)

    @ipfsOp
    async def openWithSystemDefault(self, ipfsop, rscPath):
        # Use xdg-open or open depending on the platform
        if self.app.system == 'Linux':
            await self.openWithExternal(rscPath, ['xdg-open', '%f'])
        elif self.app.system == 'Darwin':
            await self.openWithExternal(rscPath, ['open', '%f'])


class ResourceAnalyzer(QObject):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.app = QApplication.instance()

    @ipfsOp
    async def __call__(self, ipfsop, ipfsPath):
        """
        :param IPFSPath ipfsPath
        """

        path = ipfsPath.objPath
        mHashMeta = await self.app.multihashDb.get(path)

        if mHashMeta:
            # Already have metadata for this object
            mimetype = MIMEType(mHashMeta.get('mimetype'))
            statInfo = mHashMeta.get('stat')
            return mimetype, statInfo
        else:
            mimetype = await detectMimeType(path)

            statInfo = await ipfsop.objStat(path)
            if not statInfo or not isinstance(statInfo, dict):
                log.debug('Stat failed for {path}'.format(
                    path=path))
                return mimetype, None

            await ipfsop.sleep()

            # Store retrieved information in the metadata store
            metaMtype = mimetype.type if mimetype and mimetype.valid else None
            await self.app.multihashDb.store(
                path,
                mimetype=metaMtype,
                stat=statInfo
            )

        if mimetype and mimetype.valid:
            return mimetype, statInfo

    @ipfsOp
    async def decodeQrCodes(self, ipfsop, path):
        try:
            data = await ipfsop.waitFor(
                ipfsop.client.cat(path), 12
            )

            if data is None:
                return

            # Decode the QR codes in the image if there's any
            qrDecoder = IPFSQrDecoder()
            if not qrDecoder:
                return

            return qrDecoder.decode(data)
        except aioipfs.APIError:
            pass



import asyncio
import re
import collections
import time
import os.path
import copy

from PyQt5.QtCore import (pyqtSignal, QObject)

from galacteek import log, logUser, GALACTEEK_NAME, ensure

from galacteek.ipfs import pinning
from galacteek.ipfs.wrappers import ipfsOp
from galacteek.ipfs.pubsub.messages import ChatRoomMessage
from galacteek.ipfs.pubsub.service import PSMainService
from galacteek.ipfs.pubsub.service import PSChatService
from galacteek.ipfs.pubsub.service import PSHashmarksExchanger
from galacteek.ipfs.pubsub.service import PSPeersService

from galacteek.ipfs.ipfsops import *
from galacteek.ipfs import tunnel

from galacteek.core.profile import UserProfile
from galacteek.core.softident import gSoftIdent
from galacteek.crypto.rsa import RSAExecutor


class PeerCtx(QObject):
    def __init__(self, ipfsCtx, peerId, identMsg, pinglast=0, pingavg=0):
        self._ipfsCtx = ipfsCtx
        self.peerId = peerId
        self.ident = identMsg
        self.pinglast = pinglast
        self.pingavg = pingavg
        self._identLast = int(time.time())

    @property
    def ipfsCtx(self):
        return self._ipfsCtx

    @property
    def ident(self):
        return self._ident

    @property
    def identLast(self):
        return self._identLast

    @property
    def peerUnresponsive(self):
        return (int(time.time()) - self.identLast) > (60 * 10)

    @ident.setter
    def ident(self, v):
        self._ident = v
        self._identLast = int(time.time())
        self.debug('Updated last ident to {}'.format(self.identLast))

    def debug(self, msg):
        log.debug('Peer {0}: {1}'.format(self.peerId, msg))

    @ipfsOp
    async def update(self, op):
        pass

    @ipfsOp
    async def getRsaPubKey(self, op):
        pubKeyPayload = self.ident.rsaPubKeyPem

        if pubKeyPayload:
            pubKey = await self.ipfsCtx.rsaExec.importKey(
                pubKeyPayload)
            return pubKey
        else:
            self.debug('Failed to load pubkey')


class Peers(QObject):
    changed = pyqtSignal()
    peerAdded = pyqtSignal(str)
    peerModified = pyqtSignal(str)
    peerLogout = pyqtSignal(str)

    def __init__(self, ctx):
        super().__init__(ctx)

        self.ctx = ctx
        self.lock = asyncio.Lock()
        self.evStopWatcher = asyncio.Event()
        self._byPeerId = collections.OrderedDict()

    @property
    def byPeerId(self):
        return self._byPeerId

    @property
    def peersIds(self):
        return self.byPeerId.keys()

    @property
    def peersCount(self):
        return len(self.peersIds)

    @property
    def peersNicknames(self):
        return [peer.ident.username for pid, peer in self.byPeerId.items()]

    async def unregister(self, peerId):
        with await self.lock:
            if peerId in self.byPeerId:
                del self.byPeerId[peerId]
            self.peerLogout.emit(peerId)
            self.changed.emit()

    @ipfsOp
    async def registerFromIdent(self, op, iMsg):
        # identMsg is a PeerIdentMessage

        if iMsg.peer not in self.byPeerId:
            now = int(time.time())
            avgPing = await op.waitFor(op.pingAvg(iMsg.peer, count=2), 5)

            with await self.lock:
                pCtx = PeerCtx(self.ctx, iMsg.peer, iMsg,
                               pingavg=avgPing if avgPing else 0,
                               pinglast=now if avgPing else 0
                               )
                self._byPeerId[iMsg.peer] = pCtx
            self.peerAdded.emit(iMsg.peer)
        else:
            with await self.lock:
                self.peerModified.emit(iMsg.peer)
                self._byPeerId[iMsg.peer].ident = iMsg

        self.changed.emit()

    async def init(self):
        pass

    async def watch(self):
        while True:
            await asyncio.sleep(1800)

            with await self.lock:
                log.debug('Peers: clearing unresponsive peers')
                peersList = copy.copy(self.byPeerId)
                for peerId, pCtx in peersList.items():
                    if pCtx.peerUnresponsive:
                        log.debug('{} unresponsive ..'.format(peerId))
                        await self.unregister(peerId)
                del peersList

    async def stop(self):
        pass

    def getByPeerId(self, peerId):
        return self._byPeerId.get(peerId, None)

    def peerRegistered(self, peerId):
        return peerId in self.peersIds

    def __str__(self):
        return 'Galacteek peers registered: {0}'.format(self.peersCount)


class Node(QObject):
    def __init__(self, parent):
        super().__init__(parent)
        self.ctx = parent
        self._id = None
        self._idFull = None

    @property
    def id(self):
        return self._id

    @property
    def idAll(self):
        return self._idFull

    @ipfsOp
    async def init(self, op):
        self._idFull = await op.client.core.id()
        self._id = await op.nodeId()


class PubsubMaster(QObject):
    psMessageRx = pyqtSignal()
    psMessageTx = pyqtSignal()

    chatRoomMessageReceived = pyqtSignal(ChatRoomMessage)

    def __init__(self, parent):
        super().__init__(parent)
        self.ctx = parent
        self._services = {}

    @property
    def services(self):
        return self._services

    def reg(self, service):
        self._services[service.topic] = service

    def status(self):
        [service.logStatus() for topic, service in self.services.items()]

    async def send(self, topic, message):
        if topic in self.services:
            await self.services[topic].send(str(message))

    async def stop(self):
        tsks = [service.stop() for topic, service in self.services.items()]
        return await asyncio.gather(*tsks)

    def startServices(self):
        [service.start() for topic, service in self.services.items()]

    @ipfsOp
    async def init(self, op):
        pass


class P2PServices(QObject):
    def __init__(self, parent):
        super().__init__(parent)
        self.ctx = parent
        self._manager = tunnel.P2PTunnelsManager()
        self._services = []

    @property
    def services(self):
        return self._services

    def servicesFormatted(self):
        return [{
            'name': srv.name,
            'descr': srv.description,
            'protocol': srv.protocolName
        } for srv in self.services]

    async def listeners(self):
        return await self._manager.getListeners()

    async def streamsAll(self):
        return await self._manager.streams()

    def register(self, service):
        service.manager = self._manager
        self._services.append(service)

    @ipfsOp
    async def init(self, op):
        if not await op.hasCommand('p2p') is True:
            log.debug('No P2P streams support')
            return

        log.debug('P2P streams support available')

    async def stop(self):
        for srv in self.services:
            await srv.stop()


class IPFSContext(QObject):
    # signals
    ipfsConnectionReady = pyqtSignal()
    ipfsRepositoryReady = pyqtSignal()

    # profiles
    profilesAvailable = pyqtSignal(list)
    profileChanged = pyqtSignal(str, UserProfile)
    profileInfoAvailable = pyqtSignal()

    # log events
    logAddProvider = pyqtSignal(dict)

    # pubsub
    pubsubMessageRx = pyqtSignal()
    pubsubMessageTx = pyqtSignal()

    pubsubMarksReceived = pyqtSignal(int)

    # pinning signals
    pinQueueSizeChanged = pyqtSignal(int)
    pinItemStatusChanged = pyqtSignal(str, str, dict)
    pinItemRemoved = pyqtSignal(str, str)
    pinItemsCount = pyqtSignal(int)
    pinNewItem = pyqtSignal(str)
    pinFinished = pyqtSignal(str)

    def __init__(self, app):
        super().__init__()

        self._app = app

        self.objectStats = {}
        self.resources = {}
        self.profiles = {}
        self._currentProfile = None
        self.ipfsClient = None
        self._softIdent = None

        self.peers = Peers(self)
        self.node = Node(self)
        self.p2p = P2PServices(self)
        self.pubsub = PubsubMaster(self)

        self.pinner = None
        self.pinnerTask = None
        self.orbitConnector = None

    @property
    def app(self):
        return self._app

    @property
    def inOrbit(self):
        # We in orbit yet ?
        if self.orbitConnector:
            return self.orbitConnector.connected
        return False

    @property
    def loop(self):
        return self.app.loop

    @property
    def client(self):
        return self.ipfsClient

    @property
    def currentProfile(self):
        return self._currentProfile

    @currentProfile.setter
    def currentProfile(self, p):
        log.debug('Switched profile: {}'.format(p.name))
        self._currentProfile = p

    @property
    def rsaAgent(self):
        if self.currentProfile:
            return self.currentProfile.rsaAgent

    @property
    def softIdent(self):
        return self._softIdent

    @softIdent.setter
    def softIdent(self, ident):
        msg = 'Software ident changed to: CID {}'.format(ident['Hash'])
        log.debug(msg)
        logUser.info(msg)
        self._softIdent = ident

    def hasRsc(self, name):
        return name in self.resources

    @ipfsOp
    async def setup(self, ipfsop, pubsubEnable=True,
                    pubsubHashmarksExch=False, p2pEnable=True):
        self.rsaExec = RSAExecutor(loop=self.loop,
                                   executor=self.app.executor)
        await self.importSoftIdent()

        await self.node.init()
        await self.peers.init()

        if p2pEnable is True:
            await self.p2p.init()

        ensure(self.peers.watch())

        self.pinner = pinning.PinningMaster(
            self, statusFilePath=self.app.pinStatusLocation)
        await self.pinner.start()

        if pubsubEnable is True:
            self.setupPubsub(pubsubHashmarksExch=pubsubHashmarksExch)

    async def shutdown(self):
        if self.pinner:
            await self.pinner.stop()

        await self.peers.stop()
        await self.p2p.stop()
        await self.pubsub.stop()

    def setupPubsub(self, pubsubHashmarksExch=False):
        psServiceMain = PSMainService(self, self.app.ipfsClient)
        self.pubsub.reg(psServiceMain)

        psServicePeers = PSPeersService(self, self.app.ipfsClient)
        self.pubsub.reg(psServicePeers)

        psServiceChat = PSChatService(self, self.app.ipfsClient)
        self.pubsub.reg(psServiceChat)

        if pubsubHashmarksExch:
            psServiceMarks = PSHashmarksExchanger(
                self,
                self.app.ipfsClient,
                self.app.marksLocal,
                self.app.marksNetwork
            )
            self.pubsub.reg(psServiceMarks)

        self.pubsub.startServices()
        self.pubsub.status()

    @ipfsOp
    async def profilesInit(self, ipfsop):
        hasGalacteek = await ipfsop.filesLookup('/', GALACTEEK_NAME)
        if not hasGalacteek:
            await ipfsop.client.files.mkdir(GFILES_ROOT_PATH, parents=True)

        rootList = await ipfsop.filesList(GFILES_ROOT_PATH)

        # Scans existing profiles
        for entry in rootList:
            name = entry['Name']
            if entry['Type'] == 1 and name.startswith('profile.'):
                ma = re.search(r'profile\.([a-zA-Z\.\_\-]*)$', name)
                if ma:
                    profileName = ma.group(1).rstrip()
                    await self.profileNew(profileName)

        defaultProfile = 'default'

        # Create default profile if not found
        if defaultProfile not in self.profiles:
            await self.profileNew(defaultProfile)

        self.profileEmitAvail()
        self.profileChange(defaultProfile)

    def profileGet(self, name):
        return self.profiles.get(name, None)

    def profileEmitAvail(self):
        self.loop.call_soon(self.profilesAvailable.emit,
                            list(self.profiles.keys()))

    async def profileNew(self, pName, emitavail=False):
        profile = UserProfile(self, pName, os.path.join(
            GFILES_ROOT_PATH, 'profile.{}'.format(pName)))
        try:
            await profile.init()
        except Exception as e:
            log.info('Could not initialize profile: {}'.format(
                str(e)), exc_info=True)
            return None
        self.profiles[pName] = profile
        if emitavail is True:
            self.profileEmitAvail()
        return profile

    def profileChange(self, pName):
        if pName in self.profiles:
            self.currentProfile = self.profiles[pName]
            self.profileChanged.emit(pName, self.currentProfile)
            return True
        else:
            return False

    @ipfsOp
    async def importSoftIdent(self, op):
        added = await op.client.add_str(gSoftIdent)
        if added is not None:
            self.softIdent = added

    @ipfsOp
    async def galacteekPeers(self, op):
        if self.softIdent is not None:
            return await op.whoProvides(self.softIdent['Hash'])
        else:
            return []

    async def pin(self, path, recursive=False, callback=None, qname='default'):
        if self.pinner:
            await self.pinner.queue(path, recursive, callback, qname=qname)



from sys import version_info


def isoformat(dt, sep=' ', timespec='seconds'):
    if version_info.major == 3 and version_info.minor < 6:
        return dt.isoformat(sep)
    elif version_info.major == 3 and version_info.minor >= 6:
        return dt.isoformat(sep, timespec=timespec)



from galacteek.ipfs.mimetype import MIMEType


def test_mime_valid():
    mType = MIMEType('text/plain')
    assert mType.valid is True
    assert mType.category == 'text'

    mType = MIMEType('text/html; charset=UTF-8')
    assert mType.valid is True

    mType = MIMEType('image')
    assert mType.valid is False



import io

from PIL import Image

from galacteek import log
from galacteek.ipfs.cidhelpers import IPFSPath


try:
    from pyzbar.pyzbar import decode as zbar_decode
    from pyzbar import zbar_library
    zbar_library.load()
except ImportError:
    haveZbar = False
else:
    haveZbar = True


try:
    from qreader.decoder import QRDecoder
    from qreader.scanner import ImageScanner
except ImportError:
    haveQReader = False
else:
    haveQReader = True


class ImageReader:
    def _getImage(self, data):
        try:
            image = Image.open(io.BytesIO(data))
        except:
            log.debug('Cannot open image with Pillow')
            return None
        else:
            return image


class ZbarIPFSQrDecoder(ImageReader):
    def decode(self, data):
        if not isinstance(data, bytes):
            raise Exception('Need bytes')

        try:
            image = self._getImage(data)

            if image is None:
                return

            objects = zbar_decode(image)

            urls = []
            for obj in objects:
                if not isinstance(obj.data, bytes):
                    continue
                try:
                    decoded = obj.data.decode()
                except BaseException:
                    continue

                if len(decoded) not in range(1, 1024):
                    continue

                path = IPFSPath(decoded)
                if path.valid and path not in urls:
                    urls.append(path)

            if len(urls) > 0:  # don't return empty list
                return urls
        except Exception:
            return None


class QReaderIPFSQrDecoder(ImageReader):
    """
    Decodes IPFS QR codes with the qreader library
    """

    def decode(self, data):
        """
        Try to decode all the QR codes with IPFS addresses contained
        in an image. Returns a list of IPFS paths, or None if no URL
        was found.

        :param bytes data: Raw image data
        :rtype: list
        """
        if not isinstance(data, bytes):
            raise Exception('Need bytes')

        try:
            image = self._getImage(data)

            if image is None:
                return

            imgScanner = ImageScanner(image)
            results = QRDecoder(imgScanner).get_all()

            urls = []
            for obj in results:
                if not isinstance(obj, str):
                    continue
                if len(obj) not in range(1, 1024):
                    continue

                path = IPFSPath(obj)
                if path.valid and path not in urls:
                    urls.append(path)

            if len(urls) > 0:  # don't return empty list
                return urls
        except Exception:
            return None


def IPFSQrDecoder():
    # Returns a suitable QR decoder depending on the available libs
    if haveZbar:
        log.debug('Using pyzbar decoder')
        return ZbarIPFSQrDecoder()
    elif haveQReader:
        log.debug('Using qreader decoder')
        return QReaderIPFSQrDecoder()
    else:
        return None



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'galacteek/ui/ipfssearchbrowser.ui'
#
# Created by: PyQt5 UI code generator 5.10
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(496, 373)
        self.verticalLayout_2 = QtWidgets.QVBoxLayout(Form)
        self.verticalLayout_2.setObjectName("verticalLayout_2")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.webView = QtWebEngineWidgets.QWebEngineView(Form)
        self.webView.setUrl(QtCore.QUrl("about:blank"))
        self.webView.setObjectName("webView")
        self.verticalLayout.addWidget(self.webView)
        self.verticalLayout_2.addLayout(self.verticalLayout)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))

from PyQt5 import QtWebEngineWidgets



import json
import time
import sys
import collections
import copy
import re
from datetime import datetime

import aiofiles

from galacteek import log
from galacteek.core import isoformat

from PyQt5.QtCore import pyqtSignal, QObject


class MarksEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, IPFSHashMark):
            return obj.data
        return json.JSONEncoder.default(self, obj)


marksKey = '_marks'


def rSlash(path):
    return path.rstrip('/')


class IPFSHashMark(collections.UserDict):
    @property
    def markData(self):
        return self.data[self.path]

    def addTags(self, tags):
        self.data[self.path]['tags'] += tags

    def dump(self):
        print(json.dumps(self.data, indent=4))

    @staticmethod
    def fromJson(data):
        return IPFSHashMark(data)

    @staticmethod
    def make(path, title=None, datecreated=None, share=False, tags=[],
             description='', comment='', datasize=None, cumulativesize=None,
             numlinks=None, icon=None, pinSingle=False, pinRecursive=False):
        if datecreated is None:
            datecreated = isoformat(datetime.now())

        path = rSlash(path)

        mData = IPFSHashMark({
            path: {
                'metadata': {
                    'title': title,
                    'description': description,
                    'datasize': datasize,
                    'cumulativesize': cumulativesize,
                    'numlinks': numlinks,
                },
                'pin': {
                    'single': pinSingle,
                    'recursive': pinRecursive,
                    'filters': []
                },
                'datecreated': datecreated,
                'tscreated': int(time.time()),
                'comment': comment,
                'icon': icon,
                'share': share,
                'tags': tags,
            }
        })

        mData.path = path
        return mData


class IPFSMarks(QObject):
    changed = pyqtSignal()
    markDeleted = pyqtSignal(str)
    markAdded = pyqtSignal(str, dict)
    feedMarkAdded = pyqtSignal(str, IPFSHashMark)

    def __init__(self, path, parent=None, data=None, autosave=True):
        super().__init__(parent)

        self._path = path
        self._autosave = autosave
        self._marks = data if data else self.load()
        self.changed.connect(self.onChanged)
        self.lastsaved = time.time()
        self.changed.emit()

    @property
    def path(self):
        return self._path

    @property
    def autosave(self):
        return self._autosave

    @property
    def root(self):
        return self._marks

    @property
    def _root(self):
        return self._marks

    @property
    def _rootMarks(self):
        return self._root['ipfsmarks']

    @property
    def _rootCategories(self):
        return self._rootMarks['categories']

    @property
    def _rootFeeds(self):
        return self._root['feeds']

    def skeleton(self):
        return {
            'ipfsmarks': {
                'categories': {}
            },
            'feeds': {}
        }

    def load(self):
        if not self.path:
            return self.skeleton()
        try:
            with open(self.path, 'rt') as fd:
                marks = json.load(fd)
            return marks
        except Exception:
            marks = collections.OrderedDict()

        if 'ipfsmarks' not in marks:
            marks = self.skeleton()

        return marks

    def onChanged(self):
        if self.autosave is True:
            self.save()

    def save(self):
        """ Save synchronously """
        if not self.path:  # don't save
            return
        try:
            with open(self.path, 'w+t') as fd:
                self.serialize(fd)
                self.lastsaved = time.time()
        except BaseException:
            log.debug('Could not save hashmarks ({0}'.format(
                self.path))

    async def saveAsync(self):
        async with aiofiles.open(self.path, 'w+t') as fd:
            await fd.write(json.dumps(self._root, indent=4, cls=MarksEncoder))
            self.lastsaved = time.time()

    def hasCategory(self, category, parent=None):
        if parent is None:
            parent = self._rootCategories
        return category in parent

    def enterCategory(self, section, create=False):
        comps = section.lstrip('/').rstrip('/').split('/')
        return self.walk(comps, create=create)

    def walk(self, path, create=True):
        """
        Walk to a category and create intermediary parents if create
        is True. path is a list of category components
        e.g ['general', 'news'] for category path /general/news
        """
        def _walk(path, parent=None):
            for p in path:
                if p.startswith('_'):
                    return
                if p in parent.keys():
                    parent = parent[p]
                elif p not in parent.keys() and create is True:
                    self.addCategory(p, parent=parent)
                    parent = parent[p]
                else:
                    return
            return parent
        return _walk(path, parent=self._rootCategories)

    def getCategories(self):
        def _list(path, parent=None):
            for p in parent.keys():
                if p.startswith('_'):
                    continue

                fullPath = path + [p]
                yield '/'.join(fullPath)
                yield from _list(fullPath, parent=parent[p])

        return list(_list([], parent=self._rootCategories))

    def addCategory(self, category, parent=None):
        if parent is None:
            parent = self._rootCategories

        if len(category) > 256:
            return None

        if not self.hasCategory(category, parent=parent):
            parent[category] = {
                marksKey: {}
            }
            self.changed.emit()
            return parent[category]

    def getCategoryMarks(self, category, usecopy=False):
        sec = self.enterCategory(category)
        if sec:
            if usecopy:
                return copy.copy(sec[marksKey])
            else:
                return sec[marksKey]

    def getAll(self, share=False):
        cats = self.getCategories()
        _all = {}
        for cat in cats:
            catMarks = self.getCategoryMarks(cat)
            for mpath, mark in catMarks.items():
                if mark['share'] == share:
                    _all[mpath] = mark
        return _all

    def searchByMetadata(self, metadata):
        if not isinstance(metadata, dict):
            raise ValueError('Metadata needs to be a dictionary')

        categories = self.getCategories()

        title = metadata.get('title', None)
        descr = metadata.get('description', None)
        path = metadata.get('path', None)

        def metaMatch(mark, field, regexp):
            try:
                if mark['metadata'][field] and re.search(
                        regexp, mark['metadata'][field]):
                    return True
            except:
                return False

        for cat in categories:
            marks = self.getCategoryMarks(cat)
            if not marks:
                continue

            for mPath, mark in marks.items():
                if 'metadata' not in mark:
                    continue

                if path and path == mPath:
                    return mPath, mark

                if title and metaMatch(mark, 'title', title):
                    return mPath, mark

                if descr and metaMatch(mark, 'description', descr):
                    return mPath, mark

        return None, None

    def search(
            self,
            bpath,
            category=None,
            tags=[],
            delete=False):
        path = rSlash(bpath)
        categories = self.getCategories()

        for cat in categories:
            if category and cat != category:
                continue

            marks = self.getCategoryMarks(cat)
            if not marks:
                continue

            if path in marks.keys():
                if delete is True:
                    del marks[path]
                    self.markDeleted.emit(path)
                    self.changed.emit()
                    return True

                tagsOk = True
                m = marks[path]
                for stag in tags:
                    if stag not in m['tags']:
                        tagsOk = False

                if tagsOk:
                    return (path, marks[path])

    def insertMark(self, mark, category):
        # Insert a mark in given category, checking of already existing mark
        # is left to the caller
        sec = self.enterCategory(category, create=True)
        if not sec:
            return False

        # Handle IPFSHashMark or tuple
        if isinstance(mark, IPFSHashMark):
            if mark.path in sec[marksKey]:
                eMark = sec[marksKey][mark.path]

                if 'icon' in mark.markData:
                    eMark['icon'] = mark.markData['icon']
                return False
            sec[marksKey].update(mark)
            self.markAdded.emit(mark.path, mark.markData)
        else:
            try:
                mPath, mData = mark
                if mPath in sec[marksKey]:
                    # Patch some fields if already exists
                    eMark = sec[marksKey][mPath]
                    if 'icon' in mData:
                        eMark['icon'] = mData['icon']
                    return False
                sec[marksKey][mPath] = mData
                self.markAdded.emit(mPath, mData)
            except Exception:
                return False

        self.changed.emit()
        return True

    def add(self, bpath, title=None, category='general', share=False, tags=[],
            description=None, icon=None, pinSingle=False, pinRecursive=False):
        if not bpath:
            return None

        path = rSlash(bpath)

        sec = self.enterCategory(category, create=True)

        if not sec:
            return False

        if self.search(path):
            # We already have stored a mark with this path
            return False

        mark = IPFSHashMark.make(path,
                                 title=title,
                                 datecreated=isoformat(datetime.now()),
                                 share=share,
                                 tags=tags,
                                 description=description,
                                 icon=icon,
                                 pinSingle=pinSingle,
                                 pinRecursive=pinRecursive
                                 )

        sec[marksKey].update(mark)
        self.changed.emit()
        self.markAdded.emit(path, mark.markData)

        return True

    def delete(self, path):
        return self.search(path, delete=True)

    def merge(self, oMarks, share=None, reset=False):
        count = 0
        for cat in oMarks.getCategories():
            marks = oMarks.getCategoryMarks(cat)
            for mark in marks.items():
                try:
                    mPath, mData = mark
                    if share is True and mData['share'] is False:
                        continue

                    if 'pin' in mData and reset:
                        mDataNew = copy.copy(mData)
                        mDataNew['pin']['single'] = False
                        mDataNew['pin']['recursive'] = False
                        mDataNew['share'] = False
                        self.insertMark((mPath, mDataNew), cat)
                    else:
                        self.insertMark(mark, cat)
                    count += 1
                except:
                    continue
        self.changed.emit()
        return count

    def follow(self, ipnsp, name, active=True, maxentries=4096,
               resolveevery=3600, share=False, autoPin=False):
        if ipnsp is None:
            return

        feedsSec = self._rootFeeds
        ipnsp = rSlash(ipnsp)

        if ipnsp in feedsSec:
            return
        feedsSec[ipnsp] = {
            'name': name,
            'active': active,
            'maxentries': maxentries,
            'resolvepolicy': 'auto',
            'resolveevery': resolveevery,
            'resolvedlast': None,
            'share': share,
            'autopin': autoPin,
            marksKey: {},
        }
        self.changed.emit()
        return feedsSec[ipnsp]

    def feedAddMark(self, ipnsp, mark):
        feeds = self._rootFeeds
        if ipnsp not in feeds:
            return False
        feed = feeds[ipnsp]
        sec = feed[marksKey]
        if mark.path in sec:
            return False

        sec.update(mark)
        self.feedMarkAdded.emit(feed['name'], mark)
        self.changed.emit()
        return True

    def getFeeds(self):
        return list(self._rootFeeds.items())

    def getFeedMarks(self, path):
        feeds = self.getFeeds()
        for fPath, fData in feeds:
            if fPath == path:
                return fData[marksKey]

    def serialize(self, fd):
        return json.dump(self._root, fd, indent=4, cls=MarksEncoder)

    def dump(self):
        print(self.serialize(sys.stdout))

    def norm(self, path):
        return path.rstrip('/')



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'galacteek/ui/browsertab.ui'
#
# Created by: PyQt5 UI code generator 5.10
#
# WARNING! All changes made in this file will be lost!

from PyQt5 import QtCore, QtGui, QtWidgets

class Ui_BrowserTabForm(object):
    def setupUi(self, BrowserTabForm):
        BrowserTabForm.setObjectName("BrowserTabForm")
        BrowserTabForm.resize(843, 444)
        self.verticalLayout_18 = QtWidgets.QVBoxLayout(BrowserTabForm)
        self.verticalLayout_18.setObjectName("verticalLayout_18")
        self.verticalLayout_17 = QtWidgets.QVBoxLayout()
        self.verticalLayout_17.setContentsMargins(-1, -1, -1, 0)
        self.verticalLayout_17.setObjectName("verticalLayout_17")
        self.horizontalLayout_7 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_7.setSizeConstraint(QtWidgets.QLayout.SetMaximumSize)
        self.horizontalLayout_7.setObjectName("horizontalLayout_7")
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setSizeConstraint(QtWidgets.QLayout.SetMaximumSize)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.hLayoutCtrl = QtWidgets.QHBoxLayout()
        self.hLayoutCtrl.setSpacing(1)
        self.hLayoutCtrl.setObjectName("hLayoutCtrl")
        self.loadIpfsButton = QtWidgets.QToolButton(BrowserTabForm)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/share/icons/ipfs-logo-128-ice.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.loadIpfsButton.setIcon(icon)
        self.loadIpfsButton.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        self.loadIpfsButton.setObjectName("loadIpfsButton")
        self.hLayoutCtrl.addWidget(self.loadIpfsButton)
        self.loadFromClipboardButton = QtWidgets.QToolButton(BrowserTabForm)
        self.loadFromClipboardButton.setText("")
        icon1 = QtGui.QIcon()
        icon1.addPixmap(QtGui.QPixmap(":/share/icons/clipboard.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.loadFromClipboardButton.setIcon(icon1)
        self.loadFromClipboardButton.setObjectName("loadFromClipboardButton")
        self.hLayoutCtrl.addWidget(self.loadFromClipboardButton)
        self.horizontalLayout.addLayout(self.hLayoutCtrl)
        self.hashmarkThisPage = QtWidgets.QToolButton(BrowserTabForm)
        self.hashmarkThisPage.setText("")
        icon2 = QtGui.QIcon()
        icon2.addPixmap(QtGui.QPixmap(":/share/icons/hashmark-black.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.hashmarkThisPage.setIcon(icon2)
        self.hashmarkThisPage.setObjectName("hashmarkThisPage")
        self.horizontalLayout.addWidget(self.hashmarkThisPage)
        self.urlZone = QtWidgets.QLineEdit(BrowserTabForm)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.urlZone.sizePolicy().hasHeightForWidth())
        self.urlZone.setSizePolicy(sizePolicy)
        self.urlZone.setMinimumSize(QtCore.QSize(400, 0))
        self.urlZone.setMaximumSize(QtCore.QSize(3000, 26))
        font = QtGui.QFont()
        font.setFamily("Sans Serif")
        font.setPointSize(12)
        font.setItalic(False)
        self.urlZone.setFont(font)
        self.urlZone.setStyleSheet("background-color: rgb(189, 236, 255);")
        self.urlZone.setMaxLength(2048)
        self.urlZone.setObjectName("urlZone")
        self.horizontalLayout.addWidget(self.urlZone)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setSpacing(1)
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.stopButton = QtWidgets.QToolButton(BrowserTabForm)
        self.stopButton.setText("")
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/share/icons/cancel.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.stopButton.setIcon(icon3)
        self.stopButton.setObjectName("stopButton")
        self.horizontalLayout_3.addWidget(self.stopButton)
        self.reloadPageButton = QtWidgets.QToolButton(BrowserTabForm)
        self.reloadPageButton.setText("")
        icon4 = QtGui.QIcon()
        icon4.addPixmap(QtGui.QPixmap(":/share/icons/refresh.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.reloadPageButton.setIcon(icon4)
        self.reloadPageButton.setObjectName("reloadPageButton")
        self.horizontalLayout_3.addWidget(self.reloadPageButton)
        spacerItem = QtWidgets.QSpacerItem(6, 20, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem)
        self.backButton = QtWidgets.QToolButton(BrowserTabForm)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.backButton.sizePolicy().hasHeightForWidth())
        self.backButton.setSizePolicy(sizePolicy)
        self.backButton.setMaximumSize(QtCore.QSize(32, 16777215))
        self.backButton.setText("")
        icon5 = QtGui.QIcon()
        icon5.addPixmap(QtGui.QPixmap(":/share/icons/left-arrow.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.backButton.setIcon(icon5)
        self.backButton.setObjectName("backButton")
        self.horizontalLayout_3.addWidget(self.backButton)
        self.forwardButton = QtWidgets.QToolButton(BrowserTabForm)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.forwardButton.sizePolicy().hasHeightForWidth())
        self.forwardButton.setSizePolicy(sizePolicy)
        self.forwardButton.setMaximumSize(QtCore.QSize(32, 16777215))
        self.forwardButton.setText("")
        icon6 = QtGui.QIcon()
        icon6.addPixmap(QtGui.QPixmap(":/share/icons/right-arrow.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.forwardButton.setIcon(icon6)
        self.forwardButton.setObjectName("forwardButton")
        self.horizontalLayout_3.addWidget(self.forwardButton)
        spacerItem1 = QtWidgets.QSpacerItem(6, 20, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem1)
        self.zoomOutButton = QtWidgets.QToolButton(BrowserTabForm)
        self.zoomOutButton.setText("")
        icon7 = QtGui.QIcon()
        icon7.addPixmap(QtGui.QPixmap(":/share/icons/zoom-out.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.zoomOutButton.setIcon(icon7)
        self.zoomOutButton.setObjectName("zoomOutButton")
        self.horizontalLayout_3.addWidget(self.zoomOutButton)
        self.zoomInButton = QtWidgets.QToolButton(BrowserTabForm)
        self.zoomInButton.setText("")
        icon8 = QtGui.QIcon()
        icon8.addPixmap(QtGui.QPixmap(":/share/icons/zoom-in.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.zoomInButton.setIcon(icon8)
        self.zoomInButton.setObjectName("zoomInButton")
        self.horizontalLayout_3.addWidget(self.zoomInButton)
        spacerItem2 = QtWidgets.QSpacerItem(6, 20, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_3.addItem(spacerItem2)
        self.pinAllButton = QtWidgets.QToolButton(BrowserTabForm)
        self.pinAllButton.setText("")
        icon9 = QtGui.QIcon()
        icon9.addPixmap(QtGui.QPixmap(":/share/icons/pin-black.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.pinAllButton.setIcon(icon9)
        self.pinAllButton.setObjectName("pinAllButton")
        self.horizontalLayout_3.addWidget(self.pinAllButton)
        self.pinToolButton = QtWidgets.QToolButton(BrowserTabForm)
        self.pinToolButton.setText("")
        self.pinToolButton.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        self.pinToolButton.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
        self.pinToolButton.setObjectName("pinToolButton")
        self.horizontalLayout_3.addWidget(self.pinToolButton)
        self.horizontalLayout.addLayout(self.horizontalLayout_3)
        self.horizontalLayout_7.addLayout(self.horizontalLayout)
        self.verticalLayout_17.addLayout(self.horizontalLayout_7)
        self.pBarBrowser = QtWidgets.QProgressBar(BrowserTabForm)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pBarBrowser.sizePolicy().hasHeightForWidth())
        self.pBarBrowser.setSizePolicy(sizePolicy)
        self.pBarBrowser.setMaximumSize(QtCore.QSize(16777215, 2))
        self.pBarBrowser.setProperty("value", 0)
        self.pBarBrowser.setTextVisible(False)
        self.pBarBrowser.setInvertedAppearance(False)
        self.pBarBrowser.setObjectName("pBarBrowser")
        self.verticalLayout_17.addWidget(self.pBarBrowser)
        self.vLayoutBrowser = QtWidgets.QVBoxLayout()
        self.vLayoutBrowser.setSizeConstraint(QtWidgets.QLayout.SetMaximumSize)
        self.vLayoutBrowser.setObjectName("vLayoutBrowser")
        self.verticalLayout_17.addLayout(self.vLayoutBrowser)
        self.verticalLayout_18.addLayout(self.verticalLayout_17)
        self.linkInfosLabel = QtWidgets.QLabel(BrowserTabForm)
        self.linkInfosLabel.setMaximumSize(QtCore.QSize(16777215, 16))
        self.linkInfosLabel.setText("")
        self.linkInfosLabel.setObjectName("linkInfosLabel")
        self.verticalLayout_18.addWidget(self.linkInfosLabel, 0, QtCore.Qt.AlignRight)

        self.retranslateUi(BrowserTabForm)
        QtCore.QMetaObject.connectSlotsByName(BrowserTabForm)

    def retranslateUi(self, BrowserTabForm):
        _translate = QtCore.QCoreApplication.translate
        BrowserTabForm.setWindowTitle(_translate("BrowserTabForm", "Form"))
        self.loadIpfsButton.setToolTip(_translate("BrowserTabForm", "<html><head/><body><p>Browse IPFS</p></body></html>"))
        self.loadIpfsButton.setText(_translate("BrowserTabForm", "..."))
        self.hashmarkThisPage.setToolTip(_translate("BrowserTabForm", "<html><head/><body><p>Hashmark this page</p></body></html>"))
        self.stopButton.setToolTip(_translate("BrowserTabForm", "<html><head/><body><p>Stop</p></body></html>"))
        self.reloadPageButton.setToolTip(_translate("BrowserTabForm", "<html><head/><body><p>Reload page</p></body></html>"))
        self.backButton.setToolTip(_translate("BrowserTabForm", "<html><head/><body><p>Previous</p></body></html>"))
        self.forwardButton.setToolTip(_translate("BrowserTabForm", "<html><head/><body><p>Next</p></body></html>"))
        self.zoomOutButton.setToolTip(_translate("BrowserTabForm", "<html><head/><body><p>Zoom out</p></body></html>"))
        self.zoomInButton.setToolTip(_translate("BrowserTabForm", "<html><head/><body><p>Zoom in</p></body></html>"))
        self.pinAllButton.setToolTip(_translate("BrowserTabForm", "<html><head/><body><p>Automatically pins visited pages</p></body></html>"))

from . import galacteek_rc



import os.path
import time
import collections

from PyQt5.QtGui import QClipboard
from PyQt5.QtCore import QObject
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtCore import QDateTime

from galacteek import ensure
from galacteek import log
from galacteek.ipfs import ipfsOp
from galacteek.ipfs.cidhelpers import IPFSPath
from galacteek.ipfs.mimetype import detectMimeType
from galacteek.ipfs.mimetype import MIMEType


class ClipboardItem(QObject):
    mimeTypeDetected = pyqtSignal(MIMEType)
    mimeTypeAvailable = pyqtSignal(MIMEType)
    mimeIconAvailable = pyqtSignal()
    statRetrieved = pyqtSignal(dict)

    def __init__(self, ipfsPath, mimeType=None, parent=None):
        super(ClipboardItem, self).__init__(parent)

        self._ipfsPath = ipfsPath
        self._mimeType = mimeType
        self._statInfo = None
        self._cid = None
        self._ts = time.time()
        self._date = QDateTime.currentDateTime()
        self._mimeIcon = None

        self.mimeTypeDetected.connect(self.setMimeType)
        self.statRetrieved.connect(self.statInfo)

    @property
    def ts(self):
        return self._ts

    @property
    def date(self):
        return self._date

    @property
    def path(self):
        # Object path (without fragment)
        return self._ipfsPath.objPath

    @property
    def fullPath(self):
        # Full path (with fragment)
        return self._ipfsPath.fullPath

    @property
    def ipfsPath(self):
        return self._ipfsPath

    @property
    def fragment(self):
        return self._ipfsPath.fragment

    @property
    def basename(self):
        return os.path.basename(self.path)

    @property
    def cid(self):
        return self._cid

    @property
    def mimeType(self):
        return self._mimeType

    @property
    def mimeCategory(self):
        if isinstance(self.mimeType, MIMEType):
            return self.mimeType.category

    @property
    def mimeIcon(self):
        return self._mimeIcon

    @mimeIcon.setter
    def mimeIcon(self, icon):
        self._mimeIcon = icon
        self.mimeIconAvailable.emit()

    @property
    def stat(self):
        return self._statInfo

    @property
    def clipInput(self):
        return self._ipfsPath.input

    @property
    def valid(self):
        return isinstance(self.mimeType, MIMEType) and self.stat is not None

    def setMimeType(self, mimeType):
        log.debug('Detected mime-type {type} for {path}'.format(
            path=self.path, type=str(mimeType)))
        self._mimeType = mimeType
        self.mimeTypeAvailable.emit(mimeType)

    def statInfo(self, stat):
        self._statInfo = stat
        self._cid = stat.get('Hash', None)

    def __str__(self):
        return self.path

    def __repr__(self):
        return '{ts} {path}: mime {mimetype}, stat {stat}'.format(
            ts=self.ts,
            path=self.path,
            mimetype=self.mimeType,
            stat=self.stat
        )


class ClipboardTracker(QObject):
    """
    Tracks the system's clipboard activity and emits signals
    depending on whether or not the clipboard contains an IPFS CID or path
    """

    clipboardPathProcessed = pyqtSignal(IPFSPath)
    clipboardHistoryChanged = pyqtSignal()  # not used anymore

    itemRegister = pyqtSignal(ClipboardItem, bool)
    itemAdded = pyqtSignal(ClipboardItem)
    itemRemoved = pyqtSignal(ClipboardItem)
    currentItemChanged = pyqtSignal(ClipboardItem)

    def __init__(self, app, clipboard, maxItems=256):
        super(ClipboardTracker, self).__init__()

        self.app = app
        self.clipboard = clipboard

        self._current = None
        self._items = collections.deque(maxlen=maxItems)
        self.itemRegister.connect(self.addItem)

        self.clipboard.changed.connect(self.onClipboardChanged)
        self.clipboardPathProcessed.connect(self.onValidPath)

    @property
    def items(self):
        return self._items

    @property
    def current(self):
        return self._current

    @current.setter
    def current(self, item):
        self._current = item
        self.currentItemChanged.emit(item)

        if item.mimeType:
            item.mimeTypeAvailable.emit(item.mimeType)

    def __len__(self):
        return len(self.items)

    def addItem(self, obj, setToCurrent):
        self._items.append(obj)

        if setToCurrent:
            self.current = obj

        self.itemAdded.emit(obj)

    def exists(self, ipfsPath):
        """
        :param IPFSPath ipfsPath: path object
        :rtype: bool
        """

        return self.getByPath(ipfsPath) is not None

    def getByPath(self, ipfsPath):
        for obj in self.items:
            if obj.ipfsPath == ipfsPath:
                return obj

    def removeItem(self, item):
        if self.exists(item.path):
            try:
                self.items.remove(item)
                self.itemRemoved.emit(item)
            except BaseException:
                return

    def onClipboardChanged(self, mode):
        text = self.clipboard.text(mode)
        self.clipboardProcess(text, clipboardMode=mode)

    def clipboardProcess(self, text, clipboardMode=None):
        """
        Process the contents of the clipboard. If it is a valid CID/path,
        emit a signal, processed by the main window for the clipboard
        manager
        """

        if not isinstance(text, str) or len(text) > 1024:
            return

        text = text.strip().rstrip('/')

        path = IPFSPath(text)
        if path.valid:
            self.clipboardPathProcessed.emit(path)

    def getHistory(self):
        return list(self.items)

    def clearHistory(self):
        for item in self._items:
            self.itemRemoved.emit(item)
        self._items.clear()

    def getCurrent(self):
        """ Returns current clipboard item """
        return self.current

    def clipboardInit(self):
        """ Used to process the clipboard's content on application's init """
        text = self.getText()
        self.clipboardProcess(text)

    def clipboardPreferredMode(self):
        return QClipboard.Selection if self.clipboard.supportsSelection() \
            else QClipboard.Clipboard

    def setText(self, text):
        """ Sets the clipboard's text content """
        self.clipboard.setText(text, self.clipboardPreferredMode())

    def getText(self):
        """ Returns clipboard's text content from the preferred source """
        return self.clipboard.text(self.clipboardPreferredMode())

    def onValidPath(self, ipfsPath):
        existing = self.getByPath(ipfsPath)

        if not existing:
            item = ClipboardItem(ipfsPath, parent=self)
            self.itemRegister.emit(item, True)
            ensure(self.scanItem(item))
        else:
            # Set existing item as current item
            self.current = existing

    @ipfsOp
    async def scanItem(self, ipfsop, cItem):
        mimetype = None
        path = cItem.path
        mHashMeta = await self.app.multihashDb.get(path)

        if mHashMeta:
            # Already have metadata for this object
            value = mHashMeta.get('mimetype')
            if value:
                mimetype = MIMEType(value)

            statInfo = mHashMeta.get('stat')
            if statInfo:
                cItem.statRetrieved.emit(statInfo)
        else:
            mimetype = await detectMimeType(path)

            statInfo = await ipfsop.objStat(path)
            if not statInfo or not isinstance(statInfo, dict):
                log.debug('Stat failed for {path}'.format(
                    path=path))
                return
            else:
                cItem.statRetrieved.emit(statInfo)

            await ipfsop.sleep()

            # Store retrieved information in the metadata store
            metadata = {}

            if mimetype and mimetype.valid:
                metadata['mimetype'] = str(mimetype)
            if statInfo and isinstance(statInfo, dict):
                metadata['stat'] = statInfo

            if len(metadata) > 0:
                await self.app.multihashDb.store(
                    path,
                    **metadata
                )

        if mimetype and mimetype.valid:
            cItem.mimeTypeDetected.emit(mimetype)



import os
import os.path
import asyncio
import json
import inspect
import collections
import uuid

import aiohttp
from yarl import URL

from PyQt5.QtCore import QObject, pyqtSignal

from galacteek import log, ensure


class OrbitConnectorProtocol(asyncio.SubprocessProtocol):
    def __init__(self):
        self._output = bytearray()
        self.evStarted = asyncio.Event()
        self.evExited = asyncio.Event()
        self.eventsQueue = asyncio.Queue(maxsize=256)

    @property
    def output(self):
        return self._output

    def pipe_data_received(self, fd, data):
        try:
            msg = data.decode().strip()
        except BaseException:
            return

        for line in msg.split('\n'):
            try:
                obj = json.loads(line)
                ensure(self.eventsQueue.put(obj))
            except BaseException:
                log.debug('Orbit connector unknown output: {0}'.format(line))
            else:
                log.debug('Orbit connector event: {0}'.format(obj))

                if 'servicestatus' in obj:
                    if obj['servicestatus'] == 'running':
                        self.evStarted.set()
                    else:
                        self.evExited.set()

        self._output.extend(data)

    def process_exited(self):
        self.evExited.set()


class NodeOrbitProcess(object):
    def __init__(
            self,
            port=3000,
            orbitDataPath='/tmp',
            ipfsApiHost='localhost',
            ipfsApiPort=5001):
        self.loop = asyncio.get_event_loop()
        self.port = port
        self.ipfsApiHost = ipfsApiHost
        self.ipfsApiPort = ipfsApiPort
        self.orbitDataPath = orbitDataPath

    async def start(self):
        here = os.path.realpath(
            os.path.abspath(
                os.path.split(
                    inspect.getfile(
                        inspect.currentframe()))[0]))
        root = os.path.join(here, '..', '..', 'galacteek-orbital-service')

        args = ['node', os.path.join(root, 'galacteek-orbital.js')]

        args += [
            '--ipfsapihost',
            self.ipfsApiHost
        ]
        args += [
            '--ipfsapiport',
            str(self.ipfsApiPort)
        ]

        args += [
            '--serviceport',
            str(self.port)
        ]

        args += [
            '--orbitdatapath',
            str(self.orbitDataPath)
        ]

        f = self.loop.subprocess_exec(
            lambda: OrbitConnectorProtocol(),
            *args,
            stdout=asyncio.subprocess.PIPE,
            stderr=None)
        self.transport, self.proto = await f
        return True

    def stop(self):
        try:
            self.transport.terminate()
            return True
        except Exception as e:
            self.terminateException = e
            return False


class ConfigMapNotifier(QObject):
    changed = pyqtSignal()


class OrbitConfigMap(collections.UserDict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._mapName = kwargs.pop('name', str(uuid.uuid4())),
        self.notifier = ConfigMapNotifier(None)

    def __str__(self):
        return json.dumps(self.data, indent=4)

    @property
    def mapName(self):
        return self._mapName

    @property
    def json(self):
        return self.data

    def hasDatabase(self, ns, dbname):
        namespaces = self.data.setdefault('namespaces', {})
        if ns in namespaces:
            return dbname in namespaces[ns]
        return False

    def newDatabase(self, ns, dbname, dbtype, load=-1, dbuuid=None,
                    write=None, address=None, create=True):
        if self.hasDatabase(ns, dbname):
            return
        namespaces = self.data.setdefault('namespaces', {})
        if ns not in namespaces:
            namespaces[ns] = {}
        namespaces[ns][dbname] = {
            'uuid': dbuuid if dbuuid else str(uuid.uuid4()),
            'address': address,
            'type': dbtype,
            'load': load,
            'write': write,
            'create': create
        }
        self.notifier.changed.emit()


class DatabaseOperator:
    def __init__(self, ns, dbname, connector, dbtype='eventlog'):
        self.ns = ns
        self.dbname = dbname
        self.connector = connector
        self.dbtype = dbtype

    def __str__(self):
        return 'Database @{0}/{1}'.format(self.ns, self.dbname)

    async def create(self):
        return await self.connector.createdb(self.ns, self.dbname, self.dbtype)

    async def list(self, *args, **kw):
        return await self.connector.list(self.ns, self.dbname, *args, **kw)

    async def putdoc(self, *args, **kw):
        return await self.connector.putdoc(self.ns, self.dbname, *args, **kw)

    async def get(self, *args, **kw):
        return await self.connector.get(self.ns, self.dbname, *args, **kw)

    async def set(self, *args, **kw):
        return await self.connector.set(self.ns, self.dbname, *args, **kw)

    async def add(self, *args, **kw):
        return await self.connector.add(self.ns, self.dbname, *args, **kw)

    async def open(self, *args, **kw):
        resp = await self.connector.request(self.connector.endpoint(
            os.path.join(self.ns, self.dbname, 'open')))
        if resp and resp['status'] == 'success':
            return True
        return False


class OrbitConnector:
    def __init__(self, orbitDataPath='/tmp',
                 servicePort=3000):
        self.servicePort = servicePort
        self.orbitDataPath = orbitDataPath
        self._connected = False
        self.eventListeners = []
        self._configMaps = {}

    @property
    def connected(self):
        return self._connected

    def registerEventListener(self, coro):
        self.eventListeners.append(coro)

    def useConfigMap(self, cfgMap):
        self._configMaps[cfgMap.mapName] = cfgMap

    async def syncConfig(self):
        pass

    async def start(self, servicePort=None):
        if self.connected:
            return True

        port = servicePort if servicePort else self.servicePort

        self.commandsQ = asyncio.Queue()
        self.servicePort = servicePort
        self.process = NodeOrbitProcess(port=port,
                                        orbitDataPath=self.orbitDataPath)
        self.baseUrl = URL.build(scheme='http', host='localhost',
                                 port=port)
        await self.process.start()

        try:
            await asyncio.wait_for(
                self.process.proto.evStarted.wait(),
                60)
        except asyncio.TimeoutError:
            log.debug('Time out while waiting for connector to rise')
            return False

        await asyncio.sleep(0.5)

        ensure(self.readEventsTask(self.process.proto))

        self._connected = True
        return True

    async def readEventsTask(self, protocol):
        while True:
            item = await protocol.eventsQueue.get()
            if not isinstance(item, dict):
                continue

            for listener in self.eventListeners:
                await listener(item)

    async def queueCommand(self, ns, dbname, cmd):
        await self.commandsQ.put({
            'namespace': ns,
            'dbname': dbname,
            'command': cmd
        })

    async def wsListenEvents(self):
        url = self.endpoint('status')
        async with aiohttp.ClientSession() as sess:
            async with sess.ws_connect(url, autoping=False) as ws:
                while True:
                    cmd = await self.commandsQ.get()
                    await ws.send_json(cmd)
                    reply = await ws.receive_json()
                    log.debug('WS ORBIT REPLY {0}'.format(reply))

    def endpoint(self, path):
        return str(self.baseUrl.join(URL(path)))

    async def stop(self):
        await self.request('disconnect')
        self.process.stop()

    async def request(self, path, method='get', params=None):
        try:
            async with aiohttp.ClientSession() as sess:
                meth = getattr(sess, method)
                if meth is None:
                    return None
                async with meth(self.endpoint(path), params=params) as resp:
                    return await resp.json()
        except BaseException:
            log.debug('Error when running request: {path}'.format(
                path=path))

    async def getconfig(self):
        return await self.request('getconfig')

    async def publicKey(self):
        resp = await self.request('publicKey')
        if resp['status'] == 'success':
            return resp['key']

    async def reconfigure(self):
        for mapName, cfgMap in self._configMaps.items():
            try:
                await self.configure(cfgMap.json)
            except BaseException:
                log.debug('Error mapping orbit config')
                continue
            else:
                log.debug('Mapped {name}'.format(name=mapName))

    async def configure(self, config):
        async with aiohttp.ClientSession() as sess:
            log.debug('Configuring orbit with {cfg}'.format(cfg=config))
            async with sess.post(self.endpoint('map'),
                                 json=config) as resp:
                reply = await resp.json()
                log.debug('Orbit service configure reply: {0}'.format(reply))

    map = configure

    async def list(self, ns, dbname, reverse=True, limit=-1):
        params = {
            'limit': limit
        }
        if reverse:
            params['reverse'] = 1

        resp = await self.request(self.endpoint(
            os.path.join(ns, dbname, 'list')),
            params=params)

        if resp['status'] == 'success':
            return resp['value']

        async with aiohttp.ClientSession() as sess:
            async with sess.get(self.endpoint(
                os.path.join(ns, dbname, 'list')),
                    params=params) as resp:
                reply = await resp.json()
                return reply

    async def query(self, ns, dbname, text):
        async with aiohttp.ClientSession() as sess:
            async with sess.get(self.endpoint(
                os.path.join(ns, dbname, 'query')),
                    params={'search': text}) as resp:
                reply = await resp.json()
                return reply

    async def get(self, ns, dbname, key):
        async with aiohttp.ClientSession() as sess:
            async with sess.get(self.endpoint(
                    os.path.join(ns, dbname, 'get')),
                    params={'key': key}) as resp:
                result = await resp.json()
                if result['status'] == 'success' and 'value' in result:
                    return result['value']

    async def add(self, ns, dbname, value):
        async with aiohttp.ClientSession() as sess:
            async with sess.post(self.endpoint(
                    os.path.join(ns, dbname, 'add')), json=value) as resp:
                reply = await resp.json()
                return reply

    async def putdoc(self, ns, dbname, data):
        async with aiohttp.ClientSession() as sess:
            async with sess.post(self.endpoint(
                    os.path.join(ns, dbname, 'put')), json=data) as resp:
                reply = await resp.json()
                return reply

    async def set(self, ns, dbname, key, value):
        async with aiohttp.ClientSession() as sess:
            async with sess.post(self.endpoint(
                    os.path.join(ns, dbname, 'set')),
                    json={'key': key, 'value': value}) as resp:
                reply = await resp.json()
                return reply

    async def createdb(self, ns, dbname, dbtype, params={}):
        resp = await self.request(self.endpoint(
            os.path.join(ns, dbname, 'createdb')), method='post')
        if resp and resp['status'] == 'success':
            return resp['address']

    def database(self, ns, dbname, dbtype='eventlog'):
        return DatabaseOperator(ns, dbname, self, dbtype=dbtype)


class GalacteekOrbitConnector(OrbitConnector):
    def __init__(self, orbitDataPath='/tmp',
                 servicePort=3000):
        super(GalacteekOrbitConnector, self).__init__(
            orbitDataPath=orbitDataPath,
            servicePort=servicePort)

        self.dbUsernames = self.database('core', 'usernames',
                                         dbtype='eventlog')

    async def start(self, servicePort=None):
        resp = await super(GalacteekOrbitConnector, self).start(
            servicePort=servicePort)
        ensure(self.dbUsernames.open())
        return resp

    async def usernamesList(self):
        resp = await self.dbUsernames.list()
        if resp:
            usernames = [elem['value']['name'] for elem in resp]
        else:
            usernames = []
        return usernames



from PyQt5.QtCore import Qt
from PyQt5.QtCore import QCoreApplication
from PyQt5.QtWidgets import QTreeView
from PyQt5.QtWidgets import QLineEdit
from PyQt5.QtWidgets import QPushButton
from PyQt5.QtWidgets import QVBoxLayout
from PyQt5.QtWidgets import QHBoxLayout
from PyQt5.QtWidgets import QLabel
from PyQt5.QtWidgets import QHeaderView
from PyQt5.QtWidgets import QToolButton
from PyQt5.QtGui import QStandardItemModel
from PyQt5.QtGui import QBrush
from PyQt5.QtGui import QColor

from galacteek.ipfs.cidhelpers import IPFSPath
from galacteek import ensure

from .widgets import GalacteekTab
from .modelhelpers import UneditableItem
from .modelhelpers import modelSearch
from .helpers import messageBox

from .i18n import iPath
from .i18n import iCidOrPath
from .i18n import iUnknown
from .i18n import iPinned
from .i18n import iPinning
from .i18n import iPin
from .i18n import iCancel
from .i18n import iInvalidInput


def iQueue():
    return QCoreApplication.translate('PinStatusWidget', 'Queue')


def iStatus():
    return QCoreApplication.translate('PinStatusWidget', 'Status')


def iNodesProcessed():
    return QCoreApplication.translate('PinStatusWidget', 'Nodes')


class PinStatusWidget(GalacteekTab):
    COL_TS = 0
    COL_QUEUE = 1
    COL_PATH = 2
    COL_STATUS = 3
    COL_PROGRESS = 4
    COL_CTRL = 5

    def __init__(self, gWindow):
        super(PinStatusWidget, self).__init__(gWindow)

        self.tree = QTreeView()
        self.tree.setObjectName('pinStatusWidget')
        self.boxLayout = QVBoxLayout()
        self.boxLayout.addWidget(self.tree)

        self.ctrlLayout = QHBoxLayout()
        self.btnPin = QPushButton(iPin())
        self.pathLabel = QLabel(iCidOrPath())
        self.pathEdit = QLineEdit()
        self.ctrlLayout.addWidget(self.pathLabel)
        self.ctrlLayout.addWidget(self.pathEdit)
        self.ctrlLayout.addWidget(self.btnPin)
        self.vLayout.addLayout(self.ctrlLayout)
        self.vLayout.addLayout(self.boxLayout)

        self.app.ipfsCtx.pinItemStatusChanged.connect(self.onPinStatusChanged)
        self.app.ipfsCtx.pinFinished.connect(self.onPinFinished)
        self.app.ipfsCtx.pinItemRemoved.connect(self.onItemRemoved)
        self.pathEdit.returnPressed.connect(self.onPathEntered)
        self.btnPin.clicked.connect(self.onPathEntered)

        self.model = QStandardItemModel()
        self.model.setHorizontalHeaderLabels(
            ['TS', iQueue(), iPath(), iStatus(), iNodesProcessed(), ''])

        self.tree.setSortingEnabled(True)
        self.tree.setModel(self.model)
        self.tree.sortByColumn(self.COL_TS, Qt.DescendingOrder)

        for col in [self.COL_QUEUE, self.COL_PATH, self.COL_PROGRESS]:
            self.tree.header().setSectionResizeMode(
                col, QHeaderView.ResizeToContents)

        self.tree.hideColumn(self.COL_TS)

    def resort(self):
        self.model.sort(self.COL_TS, Qt.DescendingOrder)

    def onPathEntered(self):
        text = self.pathEdit.text()
        self.pathEdit.clear()

        path = IPFSPath(text)
        if path.valid:
            ensure(self.app.ipfsCtx.pinner.queue(
                path.objPath, True, None))
        else:
            messageBox(iInvalidInput())

    def removeItem(self, path):
        modelSearch(self.model,
                    search=path, columns=[self.COL_PATH],
                    delete=True)

    def onItemRemoved(self, qname, path):
        self.removeItem(path)

    def findPinItems(self, path):
        ret = modelSearch(self.model,
                          search=path, columns=[self.COL_PATH])
        if len(ret) == 0:
            return None

        itemP = self.model.itemFromIndex(ret.pop())

        if not itemP:
            return None

        idxQueue = self.model.index(itemP.row(), self.COL_QUEUE,
                                    itemP.index().parent())
        idxProgress = self.model.index(itemP.row(), self.COL_PROGRESS,
                                       itemP.index().parent())
        idxStatus = self.model.index(itemP.row(), self.COL_STATUS,
                                     itemP.index().parent())
        idxC = self.model.index(itemP.row(), self.COL_CTRL,
                                itemP.index().parent())
        cancelButton = self.tree.indexWidget(idxC)

        return {
            'itemPath': itemP,
            'itemQname': self.model.itemFromIndex(idxQueue),
            'itemProgress': self.model.itemFromIndex(idxProgress),
            'itemStatus': self.model.itemFromIndex(idxStatus),
            'cancelButton': cancelButton
        }

    def onPinFinished(self, path):
        items = self.findPinItems(path)

        if items:
            items['itemStatus'].setText(iPinned())
            items['itemProgress'].setText('OK')

            color = QBrush(QColor('#c1f0c1'))
            for item in [items['itemQname'],
                         items['itemPath'], items['itemStatus'],
                         items['itemProgress']]:
                item.setBackground(color)

            if items['cancelButton']:
                items['cancelButton'].setEnabled(False)

        self.resort()

    def onCancel(self, qname, path):
        self.removeItem(path)
        self.app.ipfsCtx.pinner.cancel(qname, path)

    def onPinStatusChanged(self, qname, path, statusInfo):
        nodesProcessed = statusInfo['status'].get('Progress', iUnknown())
        items = self.findPinItems(path)

        if not items:
            btnCancel = QToolButton()
            btnCancel.setText(iCancel())
            btnCancel.clicked.connect(lambda: self.onCancel(qname, path))
            btnCancel.setFixedWidth(140)

            itemTs = UneditableItem(str(statusInfo['ts_queued']))
            itemQ = UneditableItem(qname)
            itemP = UneditableItem(path)
            itemStatus = UneditableItem(iPinning())
            itemNodes = UneditableItem(str(nodesProcessed))
            itemC = UneditableItem('')

            self.model.invisibleRootItem().appendRow(
                [itemTs, itemQ, itemP, itemStatus, itemNodes, itemC])
            idx = self.model.indexFromItem(itemC)
            self.tree.setIndexWidget(idx, btnCancel)
        else:
            items['itemProgress'].setText(str(nodesProcessed))

        self.resort()



from galacteek.ipfs.cid import make_cid

import multihash
import re
import os.path


def joinIpfs(path):
    if isinstance(path, str):
        return os.path.join('/ipfs/', path)


def joinIpns(path):
    if isinstance(path, str):
        return os.path.join('/ipns/', path)


def stripIpfs(path):
    if isinstance(path, str):
        return path.lstrip('/ipfs/')


def isIpfsPath(path):
    if isinstance(path, str):
        return ipfsRegSearchPath(path) is not None


def isIpnsPath(path):
    if isinstance(path, str):
        return ipnsRegSearchPath(path) is not None


def shortCidRepr(cid):
    cidStr = str(cid)
    if cid.version == 0:
        return '...{0}'.format(cidStr[3 * int(len(cidStr) / 5):])
    else:
        return '...{0}'.format(cidStr[4 * int(len(cidStr) / 5):])


def shortPathRepr(path):
    if isIpfsPath(path) or isIpnsPath(path):
        basename = os.path.basename(path)
        if cidValid(basename):
            cid = getCID(basename)
            return shortCidRepr(cid)
        else:
            return basename
    else:
        cid = getCID(path)
        if cid:
            return shortCidRepr(cid)
        else:
            return path


def isMultihash(hashstring):
    """
    Check if hashstring is a valid base58-encoded multihash

    :param str hashstring: the multihash to validate
    :return: if the value is a valid multihash or not
    :rtype: bool
    """
    try:
        multihash.decode(hashstring.encode('ascii'), 'base58')
        return True
    except BaseException:
        return False


def getCID(hashstring):
    try:
        return make_cid(hashstring)
    except BaseException:
        return None


def cidConvertBase32(multihash):
    """
    Convert a base58-encoded CIDv1 to base32

    :rtype: str
    """

    cid = getCID(multihash)
    if not cid or cid.version != 1:
        return None
    return cid.encode('base32').decode()


def cidValid(cidstring):
    """
    Check if cidstring is a valid IPFS CID

    :param str cidstring: the CID to validate
    :return: if the value is a valid CID or not
    :rtype: bool
    """

    c = getCID(cidstring)
    if c is None:
        return False
    if c.version == 0:
        # Ensure that we can decode the multihash
        try:  # can raise ValueError
            if multihash.decode(c.multihash):
                return True
        except BaseException:
            return False
    elif c.version == 1:
        return True
    return False

# Regexps


ipfsPathRe = re.compile(
    r'^(\s*)?(?:fs:|ipfs:|dweb:|https?://[a-zA-Z0-9:.]*)?(?P<fullpath>/ipfs/(?P<rootcid>[a-zA-Z0-9]{46,59})\/?(?P<subpath>[a-zA-Z0-9\/\.\-\_\'\\\+\(\)]{1,1024})?)(?P<fragment>\#[a-zA-Z0-9\-\_\.\/]{1,256})?$',  # noqa
    flags=re.MULTILINE)

ipfsCidRe = re.compile(
    r'^(\s*)?(?P<cid>[a-zA-Z0-9]{46,59})$', flags=re.MULTILINE)

ipnsPathRe = re.compile(
    r'^(\s*)?(?:fs:|ipfs:|dweb:|https?://[a-zA-Z0-9:.]*)?(?P<fullpath>/ipns/([a-zA-Z0-9\.\-\_]{1,128})\/?(?P<subpath>[a-zA-Z0-9\/\.\-\_\'\\\+\(\)]{1,1024})?)(?P<fragment>\#[a-zA-Z0-9\-\_\.\/]{1,256})?$',  # noqa
    flags=re.MULTILINE)


class IPFSPath:
    maxLength = 2048

    def __init__(self, input):
        if not isinstance(input, str):
            raise ValueError('path should be a string')

        self._enableBase32 = False
        self._input = input
        self._rscPath = None
        self._subPath = None
        self._fragment = None
        self._valid = self.__analyze()

    @property
    def input(self):
        return self._input

    @property
    def valid(self):
        return self._valid

    @property
    def isIpfs(self):
        return self.valid is True and isIpfsPath(self.path)

    @property
    def isIpns(self):
        return self.valid is True and isIpnsPath(self.path)

    @property
    def isIpnsRoot(self):
        return self.isIpns and self.subPath is None

    @property
    def fragment(self):
        # Return the fragment if there's any
        return self._fragment

    @property
    def path(self):
        # Return the object path (without fragment)
        return self.objPath

    @property
    def basename(self):
        if self.valid:
            return os.path.basename(self.objPath)

    @property
    def objPath(self):
        # Return the object path (without fragment)
        return self._rscPath

    @property
    def subPath(self):
        # Return the sub path
        return self._subPath

    @property
    def fullPath(self):
        if isinstance(self.fragment, str):
            return self._rscPath + self.fragment
        else:
            return self._rscPath

    def __analyze(self):
        """
        Analyze the path and returns a boolean (valid or not)

        :rtype bool
        """

        if len(self.input) > self.maxLength:
            return False

        ma = ipfsRegSearchPath(self.input)

        if ma:
            gdict = ma.groupdict()
            if 'rootcid' not in gdict or 'fullpath' not in gdict:
                return False

            cid = ma.group('rootcid')
            if not cidValid(cid):
                return False

            self._rscPath = ma.group('fullpath').rstrip('/')
            self._fragment = gdict.get('fragment')
            self._subPath = gdict.get('subpath')
            return True

        ma = ipnsRegSearchPath(self.input)
        if ma:
            gdict = ma.groupdict()
            self._rscPath = ma.group('fullpath').rstrip('/')
            self._fragment = gdict.get('fragment')
            self._subPath = gdict.get('subpath')
            return True

        ma = ipfsRegSearchCid(self.input)
        if ma:
            cidStr = ma.group('cid')
            if not cidValid(cidStr):
                return

            cidObject = getCID(cidStr)
            if cidObject.version == 1 and self._enableBase32:
                cidB32 = cidConvertBase32(cidStr)
                if cidB32:
                    path = joinIpfs(cidB32)
            else:
                path = joinIpfs(cidStr)

            self._rscPath = path
            return True

        return False

    def __str__(self):
        return self.fullPath if self.valid else 'Invalid path: {}'.format(
            self.input)

    def __eq__(self, other):
        if isinstance(other, IPFSPath) and other.valid and self.valid:
            return self.fullPath == other.fullPath
        return False

    def __repr__(self):
        return '{path} (fragment: {frag})'.format(
            path=self.fullPath,
            frag=self.fragment if self.fragment else 'No fragment')


def ipfsRegSearchPath(text):
    return ipfsPathRe.match(text)


def ipfsRegSearchCid(text):
    return ipfsCidRe.match(text)


def ipnsRegSearchPath(text):
    return ipnsPathRe.match(text)


def ipfsPathExtract(text):
    ma = ipfsRegSearchPath(text)
    if ma:
        return ma.group('fullpath')

    ma = ipnsRegSearchPath(text)
    if ma:
        return ma.group('fullpath')

    if ipfsRegSearchCid(text):
        return joinIpfs(text)



import functools

from PyQt5.QtCore import Qt
from PyQt5.QtCore import pyqtSignal
from PyQt5.QtCore import QCoreApplication

from PyQt5.QtGui import QImage
from PyQt5.QtGui import QPixmap
from PyQt5.QtGui import QPalette
from PyQt5.QtGui import QKeySequence

from PyQt5.Qt import QSizePolicy

from PyQt5.QtWidgets import QFrame
from PyQt5.QtWidgets import QLabel
from PyQt5.QtWidgets import QScrollArea
from PyQt5.QtWidgets import QHBoxLayout
from PyQt5.QtWidgets import QVBoxLayout
from PyQt5.QtWidgets import QToolButton
from PyQt5.QtWidgets import QSpacerItem
from PyQt5.QtWidgets import QLayout

from galacteek import ensure
from galacteek import logUser
from galacteek.crypto.qrcode import IPFSQrDecoder
from galacteek.ipfs.wrappers import ipfsOp

from .widgets import GalacteekTab
from .helpers import getIcon
from .helpers import messageBox
from .clipboard import iCopyToClipboard
from .i18n import iZoomIn
from .i18n import iZoomOut
from .i18n import iOpen


def iImageGotQrCodes(count):
    return QCoreApplication.translate(
        'ImageView',
        'This image contains {} valid IPFS QR code(s) !').format(
            count)


def iImageCannotLoad(path):
    return QCoreApplication.translate(
        'ImageView',
        'Cannot load image {}').format(path)


class ImageViewerTab(GalacteekTab):
    def __init__(self, imgPath, mainW):
        super().__init__(mainW)

        self.zoomIn = QToolButton()
        self.zoomIn.setIcon(getIcon('zoom-in.png'))
        self.zoomIn.setShortcut(QKeySequence('Ctrl++'))
        self.zoomIn.setToolTip(iZoomIn())
        self.zoomOut = QToolButton()
        self.zoomOut.setIcon(getIcon('zoom-out.png'))
        self.zoomOut.setShortcut(QKeySequence('Ctrl+-'))
        self.zoomOut.setToolTip(iZoomOut())

        self.fitWindow = QToolButton()
        self.fitWindow.setCheckable(True)
        self.fitWindow.setToolTip('Fit to window')
        self.fitWindow.setIcon(getIcon('expand.png'))

        layout = QHBoxLayout()
        layout.addItem(
            QSpacerItem(10, 10, QSizePolicy.Expanding, QSizePolicy.Minimum))
        layout.addWidget(self.zoomOut, 0, Qt.AlignLeft)
        layout.addWidget(self.zoomIn, 0, Qt.AlignLeft)
        layout.addWidget(self.fitWindow, 0, Qt.AlignLeft)

        self.vLayout.addLayout(layout)

        self.view = ImageView(self)
        self.view.imageLoaded.connect(self.onImageLoaded)
        self.view.qrCodesPresent.connect(self.onQrCodesListed)
        self.vLayout.addWidget(self.view)

        self.zoomIn.clicked.connect(self.view.zoomIn)
        self.zoomOut.clicked.connect(self.view.zoomOut)
        self.fitWindow.clicked.connect(
            lambda: self.view.fitWindow(self.fitWindow.isChecked()))

        ensure(self.view.showImage(imgPath))

    def onImageLoaded(self, path):
        pass

    def onQrCodesListed(self, urls):
        # Create the scroll area and fix maximum height
        # TODO: set maximum size on resize as well
        scrollArea = QScrollArea()
        scrollArea.setSizePolicy(
            QSizePolicy.Expanding, QSizePolicy.Expanding)
        scrollArea.setWidgetResizable(True)
        scrollArea.setMaximumHeight(self.height() / 3)

        frame = QFrame(scrollArea)
        frame.setSizePolicy(
            QSizePolicy.Expanding, QSizePolicy.Expanding)
        layout = QVBoxLayout()
        layout.setSizeConstraint(QLayout.SetMinAndMaxSize)

        scrollArea.setWidget(frame)

        iconOpen = getIcon('open.png')
        lbl = QLabel()
        lbl.setText(iImageGotQrCodes(len(urls)))
        lbl.setObjectName('qrCodeCountLabel')
        lbl.setStyleSheet('QLabel { font-size: 14pt; text-align: center; }')
        layout.addWidget(lbl)
        layout.addItem(
            QSpacerItem(10, 10, QSizePolicy.Minimum, QSizePolicy.Minimum))

        def onOpenItem(url):
            ensure(self.app.resourceOpener.open(url))

        for url in urls:
            hLayout = QHBoxLayout()
            hLayout.setSizeConstraint(QLayout.SetMinAndMaxSize)

            lbl = QLabel()
            lbl.setText('<b>{}</b>'.format(url))
            lbl.setStyleSheet('QLabel { font-size: 12pt }')

            clipBtn = QToolButton()
            clipBtn.setIcon(getIcon('clipboard.png'))
            clipBtn.setToolTip(iCopyToClipboard())
            clipBtn.clicked.connect(
                functools.partial(self.app.setClipboardText, str(url)))

            openBtn = QToolButton()
            openBtn.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
            openBtn.setIcon(iconOpen)
            openBtn.setText(iOpen())
            openBtn.clicked.connect(functools.partial(onOpenItem, url))

            hLayout.addWidget(lbl)
            hLayout.addWidget(openBtn)
            hLayout.addWidget(clipBtn)
            layout.addLayout(hLayout)

        frame.setLayout(layout)
        self.vLayout.addWidget(scrollArea)


class ImageView(QScrollArea):
    imageLoaded = pyqtSignal(str)
    qrCodesPresent = pyqtSignal(list)

    def __init__(self, parent=None):
        super(ImageView, self).__init__(parent)

        self.setObjectName('ImageView')
        self.qrDecoder = IPFSQrDecoder()
        self.image = QImage()
        self.labelImage = QLabel(self)
        self.setAlignment(Qt.AlignCenter)

        self.setBackgroundRole(QPalette.Dark)
        self.labelImage.setSizePolicy(
            QSizePolicy.Ignored, QSizePolicy.Ignored)
        self.setSizePolicy(
            QSizePolicy.Expanding, QSizePolicy.Expanding)

        self.labelImage.setScaledContents(True)
        self.scaleFactor = 1.0
        self.setWidget(self.labelImage)

    def fitWindow(self, fit):
        self.setWidgetResizable(fit)

        if not fit:
            self.scaleFactor = 1.0
            self.resizePixmap()

    def zoomOut(self):
        self.scaleFactor *= 0.75
        self.resizePixmap()

    def zoomIn(self):
        self.scaleFactor *= 1.25
        self.resizePixmap()

    def resizePixmap(self):
        pixmap = self.labelImage.pixmap()

        if pixmap == 0 or pixmap is None:
            return

        self.labelImage.resize(
            self.scaleFactor * pixmap.size())

    @ipfsOp
    async def showImage(self, ipfsop, imgPath):
        try:
            imgData = await ipfsop.waitFor(
                ipfsop.client.cat(imgPath), 8
            )

            if not imgData:
                raise Exception('Failed to load image')

            img = QImage()
            img.loadFromData(imgData)
            self.image = img

            self.labelImage.setPixmap(QPixmap.fromImage(self.image))
            self.labelImage.adjustSize()
            self.resizePixmap()

            self.imageLoaded.emit(imgPath)

            if self.qrDecoder:
                # See if we have any QR codes in the image
                urls = self.qrDecoder.decode(imgData)
                if urls:
                    self.qrCodesPresent.emit(urls)

        except Exception:
            logUser.debug('Failed to load image: {path}'.format(
                path=imgPath))
            messageBox(iImageCannotLoad(imgPath))



import re

from PyQt5.QtWidgets import QDialog
from PyQt5.QtWidgets import QInputDialog

from PyQt5.QtCore import QCoreApplication
from PyQt5.QtCore import QRegExp
from PyQt5.QtCore import Qt

from PyQt5.QtGui import QClipboard
from PyQt5.QtGui import QPixmap
from PyQt5.QtGui import QImage
from PyQt5.QtGui import QRegExpValidator

from galacteek import asyncify, ensure
from galacteek.core.ipfsmarks import *
from galacteek.core import countries
from galacteek.core.profile import UserInfos
from galacteek.ipfs import cidhelpers
from galacteek.ipfs.ipfsops import *
from galacteek.ipfs.wrappers import ipfsOp, ipfsOpFn

from . import ui_addhashmarkdialog
from . import ui_addfeeddialog
from . import ui_ipfscidinputdialog, ui_ipfsmultiplecidinputdialog
from . import ui_profileeditdialog
from . import ui_donatedialog
from . import ui_profilepostmessage
from .helpers import *
from .widgets import ImageWidget

from .i18n import iDoNotPin
from .i18n import iPinSingle
from .i18n import iPinRecursive
from .i18n import iNoTitleProvided


def boldLabelStyle():
    return 'QLabel { font-weight: bold; }'


class AddHashmarkDialog(QDialog):
    def __init__(
            self,
            marks,
            resource,
            title,
            description,
            stats,
            pin=False,
            pinRecursive=False,
            parent=None):
        super().__init__(parent)

        self.ipfsResource = resource
        self.marks = marks
        self.stats = stats if stats else {}
        self.iconCid = None

        self.ui = ui_addhashmarkdialog.Ui_AddHashmarkDialog()
        self.ui.setupUi(self)
        self.ui.resourceLabel.setText(self.ipfsResource)
        self.ui.resourceLabel.setStyleSheet(boldLabelStyle())
        self.ui.newCategory.textChanged.connect(self.onNewCatChanged)
        self.ui.selectIconButton.clicked.connect(self.onSelectIcon)
        self.ui.title.setText(title)
        self.iconWidget = ImageWidget()
        self.ui.layoutIcon.addWidget(self.iconWidget, 0, Qt.AlignCenter)

        self.ui.pinCombo.addItem(iDoNotPin())
        self.ui.pinCombo.addItem(iPinSingle())
        self.ui.pinCombo.addItem(iPinRecursive())

        regexp1 = QRegExp("[A-Za-z\/\_]+")  # noqa
        self.ui.newCategory.setValidator(QRegExpValidator(regexp1))

        if pin is True:
            self.ui.pinCombo.setCurrentIndex(1)
        elif pinRecursive is True:
            self.ui.pinCombo.setCurrentIndex(2)

        if isinstance(description, str):
            self.ui.description.insertPlainText(description)

        for cat in self.marks.getCategories():
            self.ui.category.addItem(cat)

    def onSelectIcon(self):
        fps = filesSelectImages()
        if len(fps) > 0:
            ensure(self.setIcon(fps.pop()))

    @ipfsOp
    async def setIcon(self, op, fp):
        entry = await op.addPath(fp, recursive=False)
        if entry:
            cid = entry['Hash']

            if await self.iconWidget.load(cid):
                self.iconCid = cid

    def onNewCatChanged(self, text):
        self.ui.category.setEnabled(len(text) == 0)

    def accept(self):
        title = self.ui.title.text()

        if len(title) == 0:
            return messageBox(iNoTitleProvided())

        share = self.ui.share.isChecked()
        newCat = self.ui.newCategory.text()
        description = self.ui.description.toPlainText()

        if len(newCat) > 0:
            category = re.sub('//', '', newCat)
        else:
            category = self.ui.category.currentText()

        pSingle = (self.ui.pinCombo.currentIndex() == 1)
        pRecursive = (self.ui.pinCombo.currentIndex() == 2)

        mark = IPFSHashMark.make(
            self.ipfsResource,
            title=title,
            share=share,
            comment=self.ui.comment.text(),
            description=description,
            tags=self.ui.tags.text().split(),
            pinSingle=pSingle,
            pinRecursive=pRecursive,
            icon=self.iconCid,
            datasize=self.stats.get(
                'DataSize',
                None),
            cumulativesize=self.stats.get(
                'CumulativeSize',
                None),
            numlinks=self.stats.get(
                'NumLinks',
                None))

        self.marks.insertMark(mark, category)
        self.done(0)


class AddFeedDialog(QDialog):
    def __init__(self, marks, resource, feedName=None, parent=None):
        super().__init__(parent)

        self.marks = marks
        self.ipfsResource = resource

        self.ui = ui_addfeeddialog.Ui_AddFeedDialog()
        self.ui.setupUi(self)
        self.ui.resourceLabel.setText(self.ipfsResource)
        self.ui.resourceLabel.setStyleSheet(boldLabelStyle())

        if isinstance(feedName, str):
            self.ui.feedName.setText(feedName)

    def accept(self):
        share = self.ui.share.isChecked()
        autoPin = self.ui.autoPin.isChecked()

        self.marks.follow(self.ipfsResource, self.ui.feedName.text(),
                          resolveevery=self.ui.resolve.value(),
                          share=share, autoPin=autoPin)
        self.done(0)


class IPFSCIDInputDialog(QDialog):
    """ Dialog for IPFS CID input and validation """

    def __init__(self, parent=None):
        super().__init__(parent)

        self.validCid = False
        self.ui = ui_ipfscidinputdialog.Ui_CIDInputDialog()
        self.ui.setupUi(self)
        self.ui.validity.setStyleSheet(boldLabelStyle())
        self.ui.cid.textChanged.connect(self.onCIDChanged)
        self.ui.clearButton.clicked.connect(self.onClearCID)

    def onClearCID(self):
        self.clearCID()

    def clearCID(self):
        """ Clears the CID input text """
        self.ui.cid.clear()

    def onCIDChanged(self, text):
        """
        When the CID input text changes we verify if it's a valid CID and
        update the validity label
        """
        if cidhelpers.cidValid(text):
            cid = self.getCID()
            if cid:
                self.ui.validity.setText(
                    'Valid CID version {}'.format(cid.version))
                self.validCid = True
            else:
                # Unlikely
                self.ui.validity.setText('Unknown CID type')
                self.validCid = False
        else:
            self.ui.validity.setText('Invalid CID')
            self.validCid = False

    def getHash(self):
        """ Returns the hash corresponding to the input, if valid """
        cid = self.getCID()
        if cid:
            return str(cid)

    def getCID(self):
        """ Returns the CID object corresponding to the input """
        try:
            return cidhelpers.getCID(self.ui.cid.text())
        except BaseException:
            return None

    def accept(self):
        if self.validCid is True:
            self.done(1)


class IPFSMultipleCIDInputDialog(QDialog):
    """ Dialog for multiple IPFS CID input """

    def __init__(self, parent=None):
        super().__init__(parent)

        self.validCid = False
        self.ui = ui_ipfsmultiplecidinputdialog.Ui_MultipleCIDInputDialog()
        self.ui.setupUi(self)
        self.ui.validity.setStyleSheet(boldLabelStyle())
        self.ui.cid.textChanged.connect(self.onCIDChanged)
        self.ui.addCidButton.clicked.connect(self.onAddCID)

    def onAddCID(self):
        if self.validCid:
            self.ui.cidList.addItem(self.ui.cid.text())
            self.ui.cid.clear()

    def onCIDChanged(self, text):
        if cidhelpers.cidValid(text):
            self.ui.validity.setText('Valid CID')
            self.validCid = True
        else:
            self.ui.validity.setText('Invalid CID')
            self.validCid = False

    def getCIDs(self):
        cids = []
        for idx in range(0, self.ui.cidList.count()):
            item = self.ui.cidList.item(idx)
            cids.append(item.text())
        return cids

    def accept(self):
        self.done(1)


class DonateDialog(QDialog):
    def __init__(self, bcAddr, parent=None):
        super().__init__(parent)

        self.ui = ui_donatedialog.Ui_DonateDialog()
        self.ui.setupUi(self)
        self.ui.okButton.clicked.connect(self.close)
        self.ui.bitcoinClip.clicked.connect(self.onBcClip)

        self._bcAddr = bcAddr
        self.ui.bitcoinAddress.setText('<b>{0}</b>'.format(self._bcAddr))
        self.setWindowTitle('Make a donation')

    def onBcClip(self):
        self.toClip(self._bcAddr)

    def toClip(self, data):
        app = QCoreApplication.instance()
        app.clipboard().setText(data, QClipboard.Selection)
        app.clipboard().setText(data, QClipboard.Clipboard)

    def close(self):
        self.done(1)

    def accept(self):
        self.done(1)


def notEmpty(v):
    return v != ''


class ProfileEditDialog(QDialog):
    genderMapping = {
        'Unspecified': UserInfos.GENDER_UNSPECIFIED,
        'Male': UserInfos.GENDER_MALE,
        'Female': UserInfos.GENDER_FEMALE
    }

    def __init__(self, profile, parent=None):
        super().__init__(parent)

        self.profile = profile
        self.countryList = countries.countryList
        self.previousUsername = self.profile.userInfo.username

        self.ui = ui_profileeditdialog.Ui_ProfileEditDialog()
        self.ui.setupUi(self)
        self.ui.tabWidget.setCurrentIndex(0)

        self.loadCountryData()

        self.ui.labelWarning.setStyleSheet('QLabel { font-weight: bold; }')
        self.ui.username.setText(self.profile.userInfo.username)
        self.ui.firstname.setText(self.profile.userInfo.firstname)
        self.ui.lastname.setText(self.profile.userInfo.lastname)
        self.ui.email.setText(self.profile.userInfo.email)
        self.ui.org.setText(self.profile.userInfo.org)
        self.ui.city.setText(self.profile.userInfo.city)
        self.ui.bio.setText(self.profile.userInfo.bio)

        for gender, gvalue in self.genderMapping.items():
            self.ui.gender.addItem(gender)
            if self.profile.userInfo.gender == gvalue:
                self.ui.gender.setCurrentText(gender)

        if notEmpty(self.profile.userInfo.countryName):
            self.ui.countryBox.setCurrentText(
                self.profile.userInfo.countryName)
        else:
            self.ui.countryBox.setCurrentText('Unspecified')

        if notEmpty(self.profile.userInfo.avatarCid):
            self.updateAvatarCid()

        self.ui.profileCryptoId.setText('<b>{0}</b>'.format(
            self.profile.userInfo.objHash))

        self.ui.changeIconButton.clicked.connect(self.changeIcon)
        self.ui.updateButton.clicked.connect(self.save)
        self.ui.cancelButton.clicked.connect(self.close)

        self.reloadIcon()

    def updateAvatarCid(self):
        self.ui.iconHash.setText('<a href="ipfs:{0}">{1}</a>'.format(
            cidhelpers.joinIpfs(self.profile.userInfo.avatarCid),
            self.profile.userInfo.avatarCid))

    def reloadIcon(self):
        ensure(self.loadIcon())

    def getCountryCode(self, name):
        for entry in self.countryList:
            if entry['name'] == name:
                return entry['code']

    def loadCountryData(self):
        for entry in self.countryList:
            self.ui.countryBox.addItem(entry['name'])

        self.ui.countryBox.addItem('Unspecified')

    @asyncify
    async def loadIcon(self):
        @ipfsOpFn
        async def load(op, cid):
            try:
                imgData = await op.client.cat(cid)
                img1 = QImage()
                img1.loadFromData(imgData)
                img = img1.scaledToWidth(256)
                self.ui.iconPixmap.setPixmap(QPixmap.fromImage(img))
            except Exception:
                messageBox('Error while loading image')

        if self.profile.userInfo.avatarCid != '':
            await load(self.profile.userInfo.avatarCid)

    def changeIcon(self):
        fps = filesSelectImages()
        if len(fps) > 0:
            ensure(self.setIcon(fps.pop()))

    @ipfsOp
    async def setIcon(self, op, fp):
        entry = await op.addPath(fp, recursive=False)
        if entry:
            self.profile.userInfo.setAvatarCid(entry['Hash'])
            self.reloadIcon()
            self.updateAvatarCid()

    def save(self):
        kw = {}
        for key in ['username', 'firstname',
                    'lastname', 'email', 'org',
                    'city']:
            val = getattr(self.ui, key).text()
            ma = re.search(r"[a-zA-Z\_\-\@\.\+\'\´0-9\s]*", val)
            if ma:
                kw[key] = val

        try:
            kw['bio'] = self.ui.bio.toPlainText()
        except:
            pass

        country = self.ui.countryBox.currentText()
        code = self.getCountryCode(country)

        if country and code:
            self.profile.userInfo.setCountryInfo(country, code)

        genderSelected = self.ui.gender.currentText()
        for gender, gvalue in self.genderMapping.items():
            if genderSelected == gender:
                kw['gender'] = gvalue

        self.profile.userInfo.setInfos(**kw)

        if self.previousUsername != self.profile.userInfo.username:
            self.profile.userInfo.usernameChanged.emit()

        self.done(1)

    def reject(self):
        self.done(0)


class ProfilePostMessageDialog(QDialog):
    def __init__(self, profile, parent=None):
        super().__init__(parent)

        self.profile = profile

        self.ui = ui_profilepostmessage.Ui_PostMessageDialog()
        self.ui.setupUi(self)

    def accept(self):
        ensure(self.post())

    async def post(self):
        msg = self.ui.message.toPlainText()
        title = self.ui.title.text()
        if not title:
            title = 'No title'

        if msg:
            ensure(self.profile.app.postMessage(title, msg))

        self.done(1)


class ChooseProgramDialog(QInputDialog):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.setWindowTitle('Choose a program')
        self.setInputMode(QInputDialog.TextInput)
        self.setLabelText(
            '''Command arguments (example: <b>mupdf %f</b>).
                <b>%f</b> is replaced with the file path''')



import asyncio
import shutil
import functools
import re

import aioipfs

from galacteek.ipfs import ipfsOpFn


try:
    import magic
except ImportError:
    haveMagic = False
else:
    haveMagic = True


# MIME types outside of the 'text' MIME category that we treat as text
MTYPES_SUPPL_TEXT = (
    'application/x-csh',
    'application/x-sh',
    'application/xml',
    'application/x-tex',
    'application/x-latex',
    'application/javascript',
    'application/json'
)


class MimeDecodeError(Exception):
    pass


def isDirMimeType(mType):
    return mType == 'application/x-directory' or mType == 'inode/directory'


def isTextMimeType(mimeType):
    mimeCategory = getMimeCategory(mimeType)
    return (mimeCategory == 'text' and mimeType != 'text/html') or \
        mimeType in MTYPES_SUPPL_TEXT


def getMimeCategory(mType):
    if isinstance(mType, str):
        splitted = mType.split('/')
        if len(splitted) == 2:
            return splitted[0]


# Bit improvised this one
mimeReg = re.compile(
    r'^(?P<category>[a-zA-Z\-]+)\/(?P<subtype>[a-zA-Z0-9\-\+\.\_]+)'
    r'\;?\s*(?P<params>[\=\-a-zA-Z0-9]*)?$')


class MIMEType(object):
    """
    Represents a MIME type
    """

    def __init__(self, mTypeText):
        if not isinstance(mTypeText, str):
            raise ValueError('Expecting a string')

        self._mType = mTypeText
        self._matched = mimeReg.match(mTypeText)

    @property
    def type(self):
        return self._mType

    @property
    def category(self):
        if self.valid:
            comps = self._decompose()
            return comps[0]

    @property
    def valid(self):
        return self._matched is not None

    def _decompose(self):
        # Decompose the MIME type in: category, subtype, optional params
        ma = self._matched
        if ma:
            return ma.group('category'), ma.group(
                'subtype'), ma.group('params')

    @property
    def isText(self):
        return (self.category == 'text' and self.type != 'text/html') or \
            self.type in MTYPES_SUPPL_TEXT

    @property
    def isHtml(self):
        return self.type == 'text/html' or self.type == 'application/xhtml+xml'

    @property
    def isDir(self):
        return self.type == 'application/x-directory' or \
            self.type == 'inode/directory'

    @property
    def isApplication(self):
        return self.category == 'application'

    @property
    def isAudio(self):
        return self.category == 'audio'

    @property
    def isVideo(self):
        return self.category == 'video'

    @property
    def isModel(self):
        return self.category == 'model'

    @property
    def isImage(self):
        return self.category == 'image'

    @property
    def isChemical(self):
        return self.category == 'chemical'

    @property
    def isMessage(self):
        return self.category == 'message'

    @property
    def isMultipart(self):
        return self.category == 'multipart'

    @property
    def isPdf(self):
        return self.type == 'application/pdf'

    def __str__(self):
        return self.type

    def __eq__(self, value):
        if isinstance(value, str):
            return self.type == value
        elif self.__class__ == value.__class__:
            return self.type == value.type


async def detectMimeTypeFromBuffer(buff):
    """
    Guess the MIME type from a bytes buffer, using either libmagic or file(1)

    Returns a MIMEType object

    :param bytes buff: buffer data
    :rtype: MIMEType
    """
    if haveMagic:
        # Use libmagic
        # Run it in an asyncio executor
        loop = asyncio.get_event_loop()
        try:
            mime = await loop.run_in_executor(
                None, functools.partial(magic.from_buffer, buff, mime=True))
        except Exception:
            return None
        else:
            return MIMEType(mime)
    elif shutil.which('file'):
        # Libmagic not available, go with good'ol file

        try:
            proc = await asyncio.create_subprocess_shell(
                'file --mime-type -',
                stdin=asyncio.subprocess.PIPE,
                stdout=asyncio.subprocess.PIPE
            )
            stdout, stderr = await proc.communicate(buff)

            out = stdout.decode().strip()
            spl = out.split(':')
        except:
            return None
        else:
            if len(spl) == 2 and spl[0] == '/dev/stdin':
                return MIMEType(spl[1].strip())
    else:
        raise MimeDecodeError('No MIME detection method available')


@ipfsOpFn
async def detectMimeType(ipfsop, rscPath, bufferSize=512, timeout=12):
    """
    Returns the MIME type of a given IPFS resource
    Uses either python-magic if available, or runs the 'file' command

    A chunk of the file is read and used to determine its MIME type

    Returns a MIMEType object

    :param ipfsop: IPFS operator instance
    :param str rscPath: IPFS CID/path
    :param int bufferSize: buffer size in bytes
    :param int timeout: operation timeout (in seconds)
    :rtype: MIMEType
    """
    try:
        buff = await ipfsop.waitFor(
            ipfsop.client.cat(rscPath, length=bufferSize), timeout
        )
    except aioipfs.APIError as err:
        if isinstance(err.message, str) and \
                err.message.lower() == 'this dag node is a directory':
            if 0:
                # This could serve later on, right now treat it
                # always as a directory
                async for obj in ipfsop.list(rscPath, resolve_type=False):
                    for entry in obj['Links']:
                        await ipfsop.sleep()
                        if entry['Name'].startswith('index.htm'):
                            return 'text/html'
            return MIMEType('inode/directory')
    else:
        if not buff:
            return None

        return await detectMimeTypeFromBuffer(buff)



from PyQt5.QtWidgets import QWidget
from PyQt5.QtGui import QTextCursor

from galacteek.ipfs import cidhelpers
from galacteek.ipfs.wrappers import ipfsOp
from galacteek.ipfs.pubsub.messages import ChatRoomMessage
from galacteek.ipfs.pubsub import TOPIC_CHAT
from galacteek import ensure

from .widgets import GalacteekTab

from . import ui_chatroom


class ChatRoomWidget(GalacteekTab):
    def __init__(self, gWindow):
        super(ChatRoomWidget, self).__init__(gWindow)

        self.chatWidget = QWidget()
        self.vLayout.addWidget(self.chatWidget)

        self.ui = ui_chatroom.Ui_Form()
        self.ui.setupUi(self.chatWidget)
        self.ui.chatLog.setOpenExternalLinks(False)
        self.ui.chatLog.setOpenLinks(False)

        self.app.ipfsCtx.pubsub.chatRoomMessageReceived.connect(
            self.onChatMessageReceived)
        self.ui.sendButton.clicked.connect(self.onSendMessage)
        self.ui.message.returnPressed.connect(self.onSendMessage)

    def onChatMessageReceived(self, message):
        self.ui.chatLog.moveCursor(QTextCursor.End, QTextCursor.MoveAnchor)
        self.ui.chatLog.insertPlainText('\n')
        self.ui.chatLog.insertHtml('<br>')

        formatted = '<p>'
        formatted += '<span style="color: black">({date})</span> '.format(
            date=message.date)
        formatted += '<span style="color: #006080">#{channel}</span> '.format(
            channel=message.channel)
        formatted += '<span style="color: #e67300">{sender}</span>: '.format(
            sender=message.sender)
        formatted += ' {message}'.format(message=message.message)
        formatted += '</p>'

        if len(message.links) > 0:
            for link in message.links:
                formatted += '<p>Link '
                formatted += '<a href="dweb:{link}">{link}<a>'.format(
                    link=link)
                formatted += '</p>'

        self.ui.chatLog.insertHtml(formatted)

        self.ui.chatLog.verticalScrollBar().setValue(
            self.ui.chatLog.verticalScrollBar().maximum())

    def onSendMessage(self):
        messageText = self.ui.message.text()
        self.ui.message.clear()
        ensure(self.sendMessage(messageText))

    @ipfsOp
    async def sendMessage(self, ipfsop, msgText):
        profile = ipfsop.ctx.currentProfile

        links = []
        words = msgText.split()
        for word in words:
            path = cidhelpers.IPFSPath(word)
            if path.valid:
                links.append(str(path))

        msg = ChatRoomMessage.make(
            profile.userInfo.username,
            ChatRoomMessage.CHANNEL_GENERAL,
            msgText,
            links=links)
        await ipfsop.ctx.pubsub.send(TOPIC_CHAT, msg)



from io import BytesIO

from galacteek.ipfs.wrappers import ipfsOp
from galacteek.core.asynclib import asyncReadFile
from galacteek import log

import aioipfs


class IpfsRSAAgent:
    """
    IPFS RSA Agent

    :param rsaExecutor: RSA Executor object
    :param pubKeyPem: PEM-encoded public key
    :param privKeyPath: Path to private RSA key
    """

    def __init__(self, rsaExecutor, pubKeyPem, privKeyPath):
        self.rsaExec = rsaExecutor
        self.pubKeyPem = pubKeyPem
        self.privKeyPath = privKeyPath

    def debug(self, msg):
        log.debug('RSA Agent: {0}'.format(msg))

    async def encrypt(self, data, pubKey):
        return await self.rsaExec.encryptData(BytesIO(data), pubKey)

    async def decrypt(self, data):
        return await self.rsaExec.decryptData(BytesIO(data),
                                              await self.__rsaReadPrivateKey())

    @ipfsOp
    async def storeSelf(self, op, data):
        """
        Encrypt some data with our pubkey and store it in IPFS
        Returns the CID of the encrypted file
        """
        try:
            encrypted = await self.encrypt(data, self.pubKeyPem)
            if encrypted is None:
                return
            entry = await op.client.add_bytes(encrypted)
            if entry:
                self.debug(
                    'storeSelf: encoded {0} bytes to {1}'.format(
                        len(data), entry['Hash'])
                )
                return entry['Hash']
        except aioipfs.APIError as err:
            self.debug('IPFS error {}'.format(err.message))

    @ipfsOp
    async def encryptToMfs(self, op, data, mfsPath):
        try:
            encrypted = await self.encrypt(data, self.pubKeyPem)
            if not encrypted:
                return False
            return await op.filesWrite(mfsPath, encrypted,
                                       create=True, truncate=True)
        except aioipfs.APIError as err:
            self.debug('IPFS error {}'.format(err.message))

    @ipfsOp
    async def decryptIpfsObject(self, op, data):
        privKey = await self.__rsaReadPrivateKey()
        try:
            decrypted = await self.rsaExec.decryptData(BytesIO(data), privKey)
            if decrypted:
                self.debug('RSA: decrypted {0} bytes'.format(
                    len(decrypted)))
                return decrypted
        except aioipfs.APIError as err:
            self.debug('decryptIpfsObject: IPFS error {}'.format(err.message))
        except Exception as e:
            self.debug('RSA: unknown error while decrypting {}'.format(str(e)))

    @ipfsOp
    async def decryptMfsFile(self, op, path):
        try:
            data = await op.client.files.read(path)
            if data is None:
                raise ValueError('Invalid file')
        except aioipfs.APIError as err:
            self.debug('decryptMfsFile failed for {0}, '
                       'IPFS error was {1}'.format(path, err.message))
        else:
            return await self.decryptIpfsObject(data)

    async def __rsaReadPrivateKey(self):
        return await asyncReadFile(self.privKeyPath)
